#if 0
NAME=`basename $0 .c`
CMDLINE="gcc -g -Wall -o $NAME $NAME.c"
echo $CMDLINE; exec $CMDLINE; exit -1
#endif
/*
 * $Id: sndlowattest.c $
 *
 * Author: Tomi Ollila <Tomi.Ollila@tfi.net>
 *
 * Created: Mon Aug 24 18:54:08 1998 too
 * Last modifyed: Mon Aug 24 20:00:01 1998 too
 *
 * HISTORY 
 * $Log: $
 */

/* NOTE: any failure exits this program */

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h> /* required on Linux for SOL_SOCKET :( */

/* Did not have some of these in my RH5.1 installation ... I hope
   these values copied from Solaris works.. if not, have to dig
   the linux ones somewhere... this is just a test code u know :) */

/* h, did not find, if Linux supports SO_RVCLOWAT at all ... :(
   have to grab kernel-source and test... */

#ifndef SO_RCVLOWAT
#define SO_RCVLOWAT xxx
#endif

/* it is hard to find a standard place where these are defined. */
extern int errno;
extern const char *const sys_errlist[];

#define WriteCS(fd, str) write(fd, str, sizeof str -1)

static int bind_listen_(int port);
static int connect_localhost_(int port);
static int accept_(int fd);

int main()
{
  int bfd = bind_listen_(13664);
  int cfd = connect_localhost_(13664);
  int afd = accept_(bfd);
  int rv;
  char buf[1024];
  
  if (bfd < 0 || cfd < 0 || afd < 0)
  {
    printf("*** bind or connect() or accept() failed: %s\n",
	   sys_errlist[errno]);
    return 0;
  }
  
  printf("bind fd = %d, connect fd = %d, accept fd = %d\n", bfd, cfd, afd);

  WriteCS(cfd, "Test Message\n"); /* hopefully this does not fail... */

  if (rv = read(bfd, buf, sizeof buf))
  {
    printf("*** read() failed: %s\n", sys_errlist[errno]);
    return 0;
  }
      
  printf("rv = %d, data: ", rv); fflush(stdout);
  write(1, buf, rv);

  printf("Setting RCVLOWAT to 1024 bytes\n");
  setRcvLoWat(afd, 1024);

  WriteCS(cfd, "Test Message 2, should not trigger select()");

  printf("Sleeping 2 secs in select()\n");
  if (!tstSelect(afd, 2))
  {
    printf("*** select() failed: %s\n", sys_errlist[errno]);
    return 0;
  }
  
  printf("Setting RCVLOWAT (back) to 1 bytes\n");
  setRcvLoWat(afd, 1);

  printf("select()ing again, should not sleep (2 sec timeout)\n");
  if (!tstSelect(afd, 2))
  {
    printf("hmm, strange, no data in readfds\n");
    return 0;
  }

  rv = read(bfd, buf, sizeof buf);

  if (rv < 0)
  {
    printf("*** read() failed: %s\n", sys_errlist[errno]);
    return 0;
  }
      
  printf("rv = %d, data: ", rv); fflush(stdout);
  write(1, buf, rv);
  
  return 0;
}

/* support routines ... */


void setRcvLoWat(int fd, int val)
{
  if (setsockopt(bfd, SOL_SOCKET, SO_RCVLOWAT, &val, sizeof val) < 0)
  {
    printf("*** setsockopt() failed: %s\n", sys_errlist[errno]);
    exit(0);
  }
}

int getRcvBuf(int fd)
{
  int val;
  
  if (getsockopt(bfd, SOL_SOCKET, SO_RCVBUF, &val, sizeof val) < 0)
  {
    printf("*** setsockopt() failed: %s\n", sys_errlist[errno]);
    exit(0);
  }
  return val;
}

int tstSelect(int fd, int sec)
{
  fd_set rfds;
  struct timeval tv;

  FD_ZERO(&rfds);
  FD_SET(fd, &rfds);

  tv.tv_sec = sec;
  tv.tv_usec = 0;
  
  if (select(fd + 1, &rfds, NULL, NULL, &tv) < 0)
  {
    printf("*** select() failed: %s\n", sys_errlist[errno]);
    exit(0);
  }
  return FD_ISSET(fd, &rfds);
}
