/*
 * $Id: connection.h $
 *
 * Author: Tomi Ollila <too@iki.fi>
 *
 * Created: Thu Jul  2 18:32:39 1998 too
 * Last modified: Sat Sep 12 18:42:50 1998 too
 *
 * HISTORY
 * $Log: $
 */

#ifndef CONNECTION_H
#define CONNECTION_H

#ifndef NBCORE_H
#include <nbcore.h>
#endif

#ifndef NB_EXTENSIONS_H
#include <netinet/in.h>
#include "nb_extensions.h"
#endif

#ifdef SHOWHDRS
#define PRTCONNID() printf("(%d)\n", conn->count)
#endif

/* We could use the same NB_IOMsg structure instead of those 4 in
   Connection structure, but it would add a lot of hassle wit copying
   fd:s and other data all over the place.
   At a later stage, when persistent connections is supported, one is
   needed in client end, and one for each server end listed
   (per connection structure). But probably I stick w/ 2 structures
   per connection endpoint. */

typedef void (* f_nbiomsg)(NB_IOMsg *);

typedef f_nbiomsg f_nh; /* alias */

typedef struct
{
#ifdef DEBUG
  int count;
#endif  
  NB_IOMsg clird;
  NB_IOMsg cliwr;	/* write data from client to server */
#if 0  
  struct {
  } cli;
#endif  
  NB_IOMsg serrd;
  NB_IOMsg serwr;	/* write data from server to client */
  union {
    struct {
      char *	bufp;
      int	state;
      /* add here flags that affect body server sends */
    } serbody;
    struct {
      char	ishtml;		/* incoming page is text/html */
      char	referer;	/* referer handling flag in HTTPGate URL */
    } serhdr;
    struct {
      char *	errstr;
      int	errval;
    } err;
  } u1;
  char * bufstart;
  int conlen;
  
} Connection;

/* NOTES for future development:
 * When support for persistent connections are added, there will be
 * one-to-many relationship between client and server connections.
 * The above structure is changed so that there is a list for server
 * connections instead of server structure embedded. The server structure
 * is something like:

typedef struct {
  Node link;
  char * destination; \* for match in format: `host[:port]' *\
  NB_IOMsg serrd;
  NB_IOMsg serwr;
  ...
  }

 * If the protocol can be "stop-and-wait" it is perhaps the most easy way
 * to implement. then only one buffer would be needed in all traffic between
 * clien and any of the servers. But nb_SendIoEventMsg() call is needed for
 * all open descriptors where data is not being written in order to determine
 * if connection in such a socket is closed. (could a SO_RCVLOWAT used to
 * ignore all incoming data that we aren't interested at that time ??
 */

/*
 * Closing of endpoints when persistent connections are supported.
 * 1) is client closes, all server connections bound are closed.
 * 2) one server disconnects when it is not sending data to the
 *    client -> only this server connection is closed.
 * 3) Server doesn't support keep-alive or doesn't tell Content-length:
 *    -> client and all server connection is closed (asap).
 * 4) Server fails on some reason and closes connection when all data is
 *    not transmitted -> close all connections.
 *
 * End of NOTES.
 */

typedef struct {
  Connection * conn;
  struct CliHdrInfo {
    struct /* headers */ {
      char * firstaccept;	/* first Accept: header */
      char * referer;
      int    refelen;
      char * host;
      int    hostlen;
      char * sosh;   /* start-of-2nd-header */
    } h;
    struct /* urlinfo */ {
      char * content; /* content ends w/ '\0', so no len used */
      char * flags;
      int    fllen;
      char * host;
      int    hlen;
      int    port;
      int    hplen; /* length of host[:port] part */
    } u;
    struct /* flags */ {
      /* flags that affect only to headers server sends */
      char   ref;
    } f;
  } c;
  NB_ResolveMsg resolvemsg;
} CliTemp;


Connection * newConnection(void);
void deleteConnection(Connection * c);

char * getDataBuffer(void);
void freeDataBuffer(char * p);


#endif /* CONNECTION_H */
