#include "project.h"

RCID("$Id: hginit.c $");
/*
 * Author: Tomi Ollila <too@iki.fi>
 *
 * Created: Wed Jul  1 16:06:14 1998 too
 * Last modified: Sat Sep 12 18:27:15 1998 too
 *
 * HISTORY
 * $Log: $
 */

#include <stdlib.h>
#include <string.h>
#include <signal.h>

#include <nbcore.h>
#include <prgutil.h>
#include <userlog.h>

#include "httpgate.h"
#include "inetsupport.h"

static const char * argtable[] = { "-port", "-daemon", NULL };
enum { ARG_PORT, ARG_DAEMON };

struct _init
{
  PRG_Args args;
  int port;
};

static int argfunc(struct _init * is, int n, int * argcp, char *** argvp);

static void init1()
{
  PRG_Args * args = (PRG_Args *)malloc(sizeof (struct _init));

  if (args == NULL)    fatal(outofmem);

  memset(args, 0, sizeof (struct _init));

  args->argtable = argtable;
  args->argfunc = argfunc;

  prg_AddArgList(args);
}

#ifdef DEBUG
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>

static void gimmeacorefile()
{
  struct rlimit rl;

  getrlimit(RLIMIT_CORE, &rl);
  rl.rlim_cur = rl.rlim_max;
  setrlimit(RLIMIT_CORE, &rl);
}
#endif

void main(int argc, char * argv[])
{
#ifdef DEBUG
  gimmeacorefile();
#endif

  P.name = argv[0];

#ifdef DEBUG
  {
    void testprg(void);
    testprg();
  }
#endif

  debugArgs();

  nb_init();

  init1();

  if (prg_ArgParse(&argc, &argv)) exit(1);

  signal(SIGPIPE, SIG_IGN);
#if 0
  nb_SetSigHandler(hgSigHandler, NULL);
  nb_SetSignals(SIGTERM, SIGHUP, 0);
#endif

  if (nb_start()) exit(1);

  nb_main(); /* never exits */

  /*
   * however, if nb_main() exits, it has failed in select().
   */
  pfatal("nbcore select() failure:");
  exit(1);
}

static void startListen(struct _init * is);

static int argfunc(struct _init * is, int n, int * argcp, char *** argvp)
{
  switch (n) {
  case -2:	/* program name */
    break;

  case ARG_PORT:
    is->port = atoi(need_next_arg(argcp, argvp));
    TRACE(5, ("is->port == %d\n", is->port));
    break;

  case ARG_DAEMON:
    TRACE(0xff, ("daemon option given... not supported yet\n"));
    break;

  case -1:	/* rest arguments (or no arguments) */
    if ((unsigned long)(is->port - 1) > 65530)
    {
      userlog(6, "Usage: %s -port <portnumber>\n",  P.name);
      return 1;
    }
    startListen(is);
    break;
  default:      /* default is not reached if programmed correctly. */
    /* fatal(IIK) */
  }
  return 0; /* success */
}

static void startListen(struct _init * is)
{
  NB_IOEventMsg * amsg = (NB_IOEventMsg *)malloc(sizeof (NB_IOEventMsg));

  P.port = is->port; /* P.port used to be local var... */

  free(is); /* ...initialization structure not needed anymore */

  if (amsg == NULL)	fatal(outofmem);

  amsg->fd = inetBind(P.port); /* this does listen() also */
  amsg->func = acceptConn;
  amsg->len = 0; /* go get event only */

  nb_SetFd(amsg->fd);
  nb_SendReadEventReq(amsg);
}
