#include "project.h"

RCID("$Id: httpgate.c $");
/*
 * Author: Tomi Ollila <too@iki.fi>
 *
 * Created: Thu Jul  2 18:31:40 1998 too
 * Last modified: Thu Sep 10 20:21:26 1998 too
 *
 * HISTORY
 * $Log: $
 */

#include "httpgates.h"

static void initialCliInput(NB_IOMsg * rmsg);

void acceptConn(NB_IOEventMsg * amsg)
{
#if 0
  int fd;
  char * cbuf;
#endif
  Connection * conn;

  TRACEF(TRCF_FUNCENTRY, ("acceptConn(amsg = %p)\n", amsg));

  /* ready to accept(2) more */
  nb_SendReadEventReq(amsg);

  conn = newConnection();
  if (conn == 0)
    return;

  if ((conn->clird.fd = inetAccept(amsg->fd)) < 0)
    goto deleteconnection;

  if ((conn->bufstart = getDataBuffer()) == NULL)
    goto deleteconnection;

  memset(conn->bufstart, 0, sizeof (CliTemp));

  ((CliTemp *)conn->bufstart)->conn = conn; /* back reference! */

  /* initially leave some space at the start for buffer data manipulation */

  conn->clird.buf = conn->bufstart + CHOFFSET;
  conn->clird.len = (DBUFSIZE - 1) - CHOFFSET;

  conn->clird.func = initialCliInput;

  nb_SendReadReq(&conn->clird);

  return;

 deleteconnection:
  deleteConnection(conn);
  return ;
}

static int cliHeader(Connection * conn, char * p);

static void initialCliInput(NB_IOMsg * rmsg)
{
  Connection * conn = F2S(rmsg, clird, Connection);
  int len = rmsg->res;
  char * p = rmsg->buf;

  TRACEF(TRCF_FUNCENTRY, ("initialCliInput(rmsg = %p)\n", rmsg));

  if (len <= 0)
  {
    deleteConnection(conn);
    return;
  }

#ifdef DEBUG
  p[len] = '\0';
#endif

  while (len--)
  {
    if (*p++ == '\n')
    {
      rmsg->res = len;
      if (cliHeader(conn, p))
	return;
    }
  }

  /*
   * Input buffer ended. Read more to it if still here (reading more headers).
   */

  if (p >= conn->bufstart + DBUFSIZE - 1)
  {
    userlog(ULOG_ERR,  "%s had more that %d bytes of header information\n",
	    "(update)", (DBUFSIZE - 1) - CHOFFSET);
    localHttpReply(conn, RM_TOOLONGHEADER);
    return;
  }

  rmsg->buf = p;
  rmsg->len = (DBUFSIZE - 1) - (p - conn->bufstart);

  nb_SendReadReq(rmsg);
}


static int checkHttpVersion(Connection * conn, char * endp);
static void cliHeaderScan2(Connection * conn);

static int cliHeader(Connection * conn, char * p)
{
  unsigned char * buf = (unsigned char *)conn->clird.buf;
  int len = (unsigned char *)p - buf;
  int i;

  TRACEF(TRCF_FUNCENTRY, ("cliHeader(conn = %p, p = %p)\n", conn, p));

  conn->clird.buf = p; /* next line */

  if (iscntrl(buf[0])) { /* check EOF case, otherwice error */
    if ((buf[0] == '\r' && buf[1] == '\n') || buf[0] == '\n')
    {
      buf[0] = AM_EOH;
      cliHeaderScan2(conn);
      return TRUE; /* eoh */
    }
    /* else */
    localHttpReply(conn, RM_TRASHEDHEADER);
    return TRUE;
  }

  i = strmapCmp(clihdrmaptable, clihdrmaptablelen, buf);

#ifdef SHOWHDRS
  if (i == CM_POST || i == CM_GET) {
    WriteCS(1, "\n*** Original Client Headers follows..."); PRTCONNID();
  }
  write(1, buf, len);
#if 0  /* some debugging info */
  printf("input index: %d, str: %s\n", i, clihdrtable[i].string);
#endif
#endif
  
  if (len >= 3) {
    buf[1] = (len >> 8); buf[2] = len % 0xff;
    buf[0] = i > 0? i: AM_IGNORE;
  }
  else
    buf[0] = AM_IGNORE0;


  /* i and buf can be reused here */
  switch (i)
  {
  case CM_POST:
    conn->conlen = -1; /* reguest for change */
    /* fall through */
  case CM_GET: /* always first.. drops everything earlier of not */
    CLITEMP(c.h.sosh) = p; /* header scanning for output starts here */
    if (buf[clihdrlen(i)] != '/')
      break; /* gets an error on later stage... after headers are parsed */
    return checkHttpVersion(conn, p - 2);

  case CM_CONLEN:
    conn->conlen = atoi(buf + clihdrlen(i));
    break;

  case CM_ACCEPT:
    if (CLITEMP(c.h.firstaccept) == NULL)
      CLITEMP(c.h.firstaccept) = buf;
    break;

  case CM_REFERER:
    p = buf + clihdrlen(i);
    if (strncasecmp(p, "http://", 7) == 0)
      p += 7;
    CLITEMP(c.h.referer) = p;
    CLITEMP(c.h.refelen) = len - (p - (char *)buf);
    break;

  case CM_HOST:
    CLITEMP(c.h.host) = buf + clihdrlen(i);
    for (p--; isspace(*p); p--)
      ;
    CLITEMP(c.h.hostlen) = p - CLITEMP(c.h.host) + 1;
    break;

  default:
    break;
  }

  /* MISC THOUGHTS...
     in above, someone interested in issue, might check how many
     entries are needed in header table that this approach is
     faster than having a linear search. For example in 8 entries
     the average amount of comparisons  is 4, w/ binary table maximun
     amount of comparisons is 3 (average is perhaps a bit over 2), but
     the search algorithm is more complicated. W/ these low enumerated
     values, compiler probably generates a jump table for that
     switch -statement. */

  return FALSE;
}

static int checkHttpVersion(Connection * conn, char * endp)
{
  /*
   * Checks that line ends w/ `HTTP/x.x\r'
   */

  TRACEF(TRCF_FUNCENTRY, ("checkHttpVersion(conn = %p, endp = %p)\n",
			  conn, endp));

  while (isspace(*endp))    endp--; /* there is at least " /\n" */
  while (!isspace(*endp))   endp--;

  if (strncasecmp(endp + 1, "HTTP/", 5) != 0 ||
      isspace(endp[6]) || isspace(endp[7]) ||isspace(endp[8]) ||
      !isspace(endp[9]))
  {
    noHttp1x(conn);
    return -1;
  }

  /* mark EOU */
  *endp = '\0';

  return FALSE;
}

/* Now we got into part where the header data is actually parsed */

/* parse host[:port]/ */

static char * getHostPort(CliTemp * ct, char * p)
{
  int port = 80;

  TRACEF(TRCF_FUNCENTRY, ("getHostPort(ct = %p, p = %p)\n", ct, p));

  ct->c.u.host = p;

  for (; *p != ':' && *p != '/' && isprint(*p); p++)
  {
    /* all allowed characters in internet host name (add more) */
    if (isalnum(*p) || *p == '.' || *p == '-' /*|| */)
      continue;

    return NULL;
  }

  ct->c.u.hlen = p - ct->c.u.host;
  /* possible port part */
  if (*p == ':')
  {
    port = atoi(++p);

    if ((unsigned int)(port - 1) > 65500)
      return NULL;

    while (isdigit(*p))
      p++;

    if (*p != '/' && isprint(*p))  /* check that port was all digits */
      return NULL;
  }

  ct->c.u.port = port;
  ct->c.u.hplen = p - ct->c.u.host;

  TRACEF(TRCF_FUNCEXIT, ("p = %p = getHostPort()\n", p));

  return p;
}

/* parse  [flags]:host[:port]/ */

static char * getFlagsHostPort(CliTemp * ct, char * p)
{
  TRACEF(TRCF_FUNCENTRY, ("getFlagsHostPort(ct = %p, p = %p)\n", ct, p));

  ct->c.u.flags = p++;

  /* flags part */
  for (; *p != ':'; p++)
  {
    switch (*p)
    {
    case 'r':  ct->c.f.ref = 0; break;
    case 'R':  ct->c.f.ref = 1; break;
/*    case 'x': ct->conn->ser.xyz =1; break;  */
    default:
      return NULL;
    }
  }

  ct->c.u.fllen = ++p - ct->c.u.flags; /* flag len includes trailing ':' */

  return getHostPort(ct, p);
}

/* viewedpage checks whether the url requested will be the page being viewed,
   or like an image in a page. This is checked by the (heuristic) fact, that
   if `Accept:' contains "*7*" (7 == /) then it is the page being viewed */


static int viewedpage(Connection * conn)
{
  char * p = CLITEMP(c.h.firstaccept);

  if (p == NULL)
    return TRUE; /* FALSE? */

  do {
    int l = (p[1] << 8) + p[2];
    char * q = p + l;

    p += clihdrlen(CM_ACCEPT);
    l -= clihdrlen(CM_ACCEPT);

    for (; l; p++, l--)
      if (p[0] == '*' && p[1] == '/' && p[2] == '*')
	return TRUE;

    p = q;
  } while (*p == CM_ACCEPT);

  return FALSE;
}

static void localPage(Connection * conn, char * url);


static void cliHeaderScan2(Connection * conn)
{
  CliTemp * ct = (CliTemp *)conn->bufstart;
  char * hdr = conn->bufstart + CHOFFSET;
  char * url = hdr + clihdrlen((unsigned char)*hdr);
  char * p;

  /* check that Host: header has been given, needed for returned
     URL host, and used to eliminate loops
     (such as http://127.0.0.1:8080/r:127.0.0.1:8080/r:127.0.0.1:8080/...)
  */
  /* though looping prevention needs more work, caseinsensitive comparisons,
     ip-address matches etc. (therefore not done here). */

  TRACEF(TRCF_FUNCENTRY, ("cliHeaderScan2(conn = %p)\n", conn));
  TRACEF(TRCF_FUNCENTRY, ("^^^^ url = `%s'\n", url));

  if (*url != '/')
  {
    localHttpReply(conn, RM_BADREQUEST);
    return;
  }

  /* at most times below, p points to start of url content */

  /* check whether got url contains [flags]:host[:port] */
  if ((p = getFlagsHostPort(ct, url)) == NULL)
  {
    /* if not, resolve destination flags, host and port from other
       headers received, such as `referer:', but first check that
       there is `host:' header so that it can be comparew w/ referer: */
    if (ct->c.h.host == NULL)
    {
      localHttpReply(conn, RM_NOHOSTHEADER);
      return;
    }

    /* if there is no referer, or host: and start of referer does not match,
     (or there is no flags:host[:port] on referer) try local page. */

    if (ct->c.h.referer == NULL ||
	strncmp(ct->c.h.referer, ct->c.h.host, ct->c.h.hostlen) || 
	ct->c.h.referer[ct->c.h.hostlen] != '/')
    {
      localPage(conn, url);
      return;
    }
    p = ct->c.h.referer + ct->c.h.hostlen;

    if (*p != '/' || !(p = getFlagsHostPort(ct, p)))
    {
      localPage(conn, url);
      return;
    }

    if (memcmp(url, "/ttp://", 7) == 0)
    {
      /* host and port set by getFlagsHostPort needs reparsing */

      if ((p = getHostPort(ct, url + 7)) == NULL)
      {
	localHttpReply(conn, RM_INVALIDURL);
	return;
      }
    }
    else
      p = url; /* urlcontent */

    if (viewedpage(conn))
    {	/* page to be viewed, do redirect */
      char * flags = ct->c.u.flags;
      int    fllen = ct->c.u.fllen; /* with tailing ':' */
      char * hopo = ct->c.u.host;
      int    hopolen = ct->c.u.hplen;
      char * q;
      char * r = q = redirect_start(conn);  /* destroys CliTmp */

      /* w/ machines that have many registers, the variable loading above
	 does not generate any overhead (would be read anyway) */

      /* copy flags */
      memcpy(q, flags, fllen);
      q += fllen;

      /* copy host[:port] */
      memcpy(q, hopo, hopolen);
      q += hopolen;

      /* copy contents */
      q = stpcpy(q, p);

      redirect(conn, r, q);
      return;
    }
  }

  ct->c.u.content = p;

#if 1
  ct->resolvemsg.msg.func = (void (*)(NB_Msg *))getConnected;

  nb_Resolve(&ct->resolvemsg, ct->c.u.host, ct->c.u.hlen);

#else  

  /* At initial coding phase this showed our resolved destination URL */
  {
    char buf[4096];
    char * p = buf + ct->c.u.hplen;

    memcpy(buf, ct->c.u.host, ct->c.u.hplen);

    CSTPCPY(p, "<p>");

    p = stpcpy(p, ct->c.u.content);

    showDataOnPage(conn, buf, p - buf);
  }
#endif

}

static void parseGoQuery(NB_IOMsg * rmsg);

static void localPage(Connection * conn, char * url)
{
  TRACEF(TRCF_FUNCENTRY, ("localPage(conn = %p, url = %s)\n", conn, url));

  url++; /* we know first char is '/' */

  if (*url == '\0')
  {
    localHttpReply(conn, RM_MAINPAGE);
    return;
  }

  if (strcmp(url, "go") == 0)
  {
    int conlen = conn->conlen;
    
    if (conlen <= 0)
    {
      /* XXX change: either a GET req or POST req w/o conlen */
      localHttpReply(conn, RM_BADREQUEST);
      return;
    }
    /* XXX here is a risk of off-by-one... check later */
    if ((char *)conn->clird.buf + conlen >= conn->bufstart + DBUFSIZE)
    {
      /* content data would overflow input buffer... */
      localHttpReply(conn, RM_TOOLONGCONTENT);
      return;
    }

#if 0 /* XXX */
    if (conlen < conn->clird.res)
    {
      /* more data received that informed !!! */
      localHttpReplyX(conn, "RM_INDSIZMISMATCH");
      return;
    }
#endif

    /* set conn->clird.len as the length of the missing part of our
       ..., so it is very easy to complete in parseGoGuery() */
    conn->clird.len = conlen;
    conn->clird.func = parseGoQuery;

    if (conn->clird.res)
      parseGoQuery(&conn->clird);
    else
      nb_SendReadReq(&conn->clird);

    return;
  }

  localHttpReply(conn, RM_NOLOCALPAGE);
}


/*********************************************************************/

static void parseGoQuery(NB_IOMsg * rmsg)
{
  Connection * conn = F2S(rmsg, clird, Connection);

  TRACEF(TRCF_FUNCENTRY, ("parseGoQuery(rmsg = %p)\n", rmsg));

  if (rmsg->res <= 0) /* EOF, possible when this function sends read request */
  {
    deleteConnection(conn);
    return;
  }

  if (rmsg->res < rmsg->len)
  {
    rmsg->buf += rmsg->res;
    rmsg->len -= rmsg->res;
    nb_SendReadReq(rmsg);
    return;
  }

#if 0
  printf("*\n* %s\n*\n", conn->cli.bufline);
#endif
  /* getting info part... */
  {
    char * bufline = conn->clird.buf;
    int    conlen = conn->conlen;
    char * q;
    char * r = q = redirect_start(conn); /* destroys CliTemp */
    int i;
    char * url = NULL;
    char ref = 'x';

    do {
      /* strstuff could be utilized here... */
      if (strncmp(bufline, "url=", 4) == 0)
	url = bufline + 4;
      else if (strncmp(bufline, "ref=", 4) == 0)
	ref = bufline[4];

      do {
	bufline++; conlen--;
      } while (*bufline != '&' && conlen > 0);

      *bufline++ = '\0'; /* having (DBUFSIZE -1) up there allows this always */
      
    } while (conlen > 0);

    if (*url == '\0' || url == NULL)
    {
      localHttpReply(conn, RM_LOCNOHTTP);
      return;
    }

    if (ref != 'x')
      *q++ = ref;

    *q++ = ':';

#if 0 /* Misthought */
    memcpy(q, host, hostlen);
    q+= hostlen;
    *q++ = '/';
#endif

    i = duecopy(q, url);

    q += i;

#if 0    
    write(1, "redir: ", 7);
    write(1, r, q - r);
#endif
    
    redirect(conn, r, q);
  }
}
