#include "project.h"

RCID("$Id: httpgate2.c $");
/*
 * Author: Tomi Ollila <Tomi.Ollila@iki.fi>
 *
 * Created: Tue Aug 25 21:06:21 1998 too
 * Last modified: Sat Sep 12 20:51:58 1998 too
 *
 * HISTORY 
 * $Log: $
 */

#include "httpgates.h"

#define sb_state u1.serbody.state
#define sb_bufp	 u1.serbody.bufp

#define sh_ishtml  u1.serhdr.ishtml
#define sh_referer u1.serhdr.referer

#if 0
/* ********** If any events from client, close... ********** */

void cliEvent(NB_IOMsg * emsg)
{
  Connection * conn = F2S(emsg, clird, Connection);

  /* XXX something to do about penging resolvemsg (check if any) */
  /* should change function pointer in below (getConnected()) */
  /*t this is a complicated matter, POST data is not read yet...*/
  /* ... means add this test later */

  /* In order to implemen this we want to know whether connection has been
     closed... but not if there is data to be read...
     There is a possibility of setting SO_RCVLOWAT value higher than
     SO_RCVBUF is. Since this does not work on my development platform
     (linux) this will be implemented when it is tested on Solaris and/or
     BSD derived systems */

  /* It might work as anticipated on those platforms either... */
  
  deleteConnection(conn);
}
#endif

/* ********** Now we start doing the real thing(tm) ********** */

static void getConnected2(NB_IOMsg * cmsg);
void getConnected(NB_ResolveMsg * rmsg)
{
  CliTemp * ct = F2S(rmsg, resolvemsg, CliTemp);
  Connection * conn = ct->conn;
  int fd, rv;
  extern int errno;

  TRACEF(TRCF_FUNCENTRY, ("getConnected(rmsg = %p)\n", rmsg));
  
  if (rmsg->addr.s_addr == 0)
  {
    localHttpReply(conn, RM_HOSTNOTFOUND);
    return;
  }
  
  if ((fd = inetSocket()) < 0)
  {
    localHttpReplySysErr(conn, "socket", errno);
    return;
  }

  conn->cliwr.fd = fd;
  conn->cliwr.func = getConnected2;
  
  if ((rv = inetConnect(fd, rmsg->addr, ct->c.u.port)) < 0)
  {
    localHttpReplySysErr(conn, "connect", errno);
    return;
  }

  /*
   * We could perhaps send a TimeMsg for connect() completion...
   * If timemsg retursn, trash connection...
   */
  
  if (rv)
    nb_SendReplyMsg((NB_Msg *)&conn->cliwr);
  else {
    conn->cliwr.len = 0; /* only event wanted for connect(2) finalization */
    nb_SendWriteEventReq((NB_IOEventMsg *)&conn->cliwr);
  }
}

static void writeRequest(Connection * conn);

static void getConnected2(NB_IOMsg * cmsg)
{
  Connection * conn = F2S(cmsg, cliwr, Connection);
  long ov;
  int  ovl = sizeof ov;
  extern int errno; /* this may be a #define on some compiler environs ?!? */
  
  TRACEF(TRCF_FUNCENTRY, ("getConnected2(cmsg = %p)\n", cmsg));

  if (getsockopt(cmsg->fd, SOL_SOCKET, SO_ERROR, &ov, &ovl) < 0)
  {
    localHttpReplySysErr(conn, "getsockopt", errno);
    return;
  }

  if (ov != 0)
  {
    ERRVAL = ov;
    localHttpReply(conn, RM_CONNFAILED);
    return;
  }

  writeRequest(conn);
  
}

/* similar code is here and in `findHeader()', should start using
   same subfunctions */

static char * writeCliHeaders(char * buf, char * header)
{
  while (1)
  {
    int hv = *header;
    int hl = (header[1] << 8) + header[2];

    switch (hv)
    {
    case AM_EOH:
      *buf++ = '\r';
      *buf++ = '\n';
      return buf;
      
    case AM_IGNORE0:
      while (*header != '\n')
	header++;
      header++;
      continue;

    case CM_USERAGENT:
    case CM_ACCLANG:
    case CM_ACCHAR:
    case CM_ACCEPT:
    case CM_CONTYPE:
    case CM_CONLEN:
      memcpy(buf, clihdrtable[hv].string, clihdrtable[hv].strlen);
      buf += clihdrtable[hv].strlen;
      memcpy(buf, header + clihdrlen(hv), hl - clihdrlen(hv));
      buf += hl - clihdrlen(hv);
      break;
      /* all others are ignored or already added */
      /* should all headers to be added here (preserving order) ? */
    }
    header += hl;
  }
  /* NOTREACHED */
  return NULL;
}

static void cliHeaderWritten(NB_IOMsg * wmsg);

static void writeRequest(Connection * conn)
{
  struct CliHdrInfo c;
  char * buf = conn->bufstart; /* destroys CliTemp */

  /* copy important data before structure destroyed */
  /* maybe this could be a floating tmp structure somewhere */
  /* similar alloc/free system than in conn structure */
  /* then copies would not be needed */
  memcpy(&c, &CLITEMP(c), sizeof c);
  
  if (conn->conlen)
    CSTPCPY(buf, "POST ");
  else
    CSTPCPY(buf, "GET ");

  /* http://localhost:<port>/flags:host[:port] is possible */
  if (*c.u.content != '/')
    *buf++ = '/';
  else
    buf = stpcpy(buf, c.u.content);

  CSTPCPY(buf, " HTTP/1.0\r\n");
  
  /* now referer checking based on flags */
  if (c.f.ref >= 0) /* here, if refer is sent at all, */
  {
    if (c.h.referer)
    {
      int l = c.h.refelen - c.h.hostlen - 1;
      char * p = c.h.referer + c.h.hostlen + 1;

      /* check that referer has [flags]:host[:port] */
      /* actually check only that there is one ':' */
      while (--l > 0)
	if (*p++ == ':')
	  break;
      if (l && (c.f.ref || strncmp(p, c.u.host, c.u.hplen) == 0))
      { /* here check if refer is sent only when referer and url matches */
	CSTPCPY(buf, "Referer: http://");
	memcpy(buf, p, l);
	buf+= l;
      }
    }
  }

  if (buf >= c.h.sosh)
  {
    localHttpReply(conn, RM_TOOLONGHEADER);
    return;
  }
  
  /* currently can handle only Connection: close */

  CSTPCPY(buf, "Connection: close\r\n");
  
  /* now check whether Host: header is sent (not in some testing case) */
  if (1)
  {
    CSTPCPY(buf, "Host: ");
    memcpy(buf, c.u.host, c.u.hplen);
    buf += c.u.hplen;
    *buf++ = '\r'; *buf++ = '\n';
  }

  if (buf >= c.h.sosh)
  {
    localHttpReply(conn, RM_TOOLONGHEADER);
    return;
  }
  
  /* now copy rest known headers from header stream */
  
  buf = writeCliHeaders(buf, c.h.sosh);

  /* in above, the headers could be redirected in original order, if
     header copy is done like server header handling below. It just needs
     to check c.h.sosh every time (at least until `referer:' is handled */
  
#ifdef SHOWHDRS
  WriteCS(1, "\n*** Modified Client headers follows..."); PRTCONNID();
  write(1, conn->bufstart, buf - conn->bufstart);
#endif

#if 1  
  conn->cliwr.buf = conn->bufstart;
  conn->cliwr.len = buf - conn->bufstart;

  conn->cliwr.func = cliHeaderWritten;

  nb_SendWriteReq(&conn->cliwr);
#endif  
}

static void initialSerInput(NB_IOMsg * rmsg);

static void readContent(NB_IOMsg * rmsg); /* past tense */
static void wroteContent(NB_IOMsg * wmsg);

static void cliHeaderWritten(NB_IOMsg * wmsg)
{
  Connection * conn = F2S(wmsg, cliwr, Connection);
  
  if (wmsg->res != wmsg->len)
  {
    localHttpReplyX(conn, "RM_XWRITEFAILED");
    return;
  }

  /*
   * Let's make server read structure ready...
   */
  conn->serrd.fd = conn->cliwr.fd;
  conn->serrd.buf = conn->bufstart + CHOFFSET;
  conn->serrd.len = DBUFSIZE - CHOFFSET;
  conn->serrd.func = initialSerInput;

  /* used temporarily for header copy destination */
  conn->serwr.buf = conn->bufstart;
  
  if (conn->conlen)
  {
    conn->cliwr.func = wroteContent;
    conn->clird.func = readContent;

    if (conn->clird.res)
    {
      readContent(&conn->clird);
      conn->clird.buf = conn->bufstart;
      conn->clird.len = DBUFSIZE; /* from now on, we don't need to write ... */
    }
    else
    {
      conn->clird.buf = conn->bufstart;
      conn->clird.len = DBUFSIZE; /* ... '\0' at the end of input buffer */
      nb_SendReadReq(&conn->clird);
    }
  }
  else /* start reading data from server... */
    nb_SendReadReq(&conn->serrd);
}

static void readContent(NB_IOMsg * rmsg)
{
  Connection * conn = F2S(rmsg, clird, Connection);

  if (rmsg->res <= 0)
  {
    /* client closed!!! userlog??? */
    deleteConnection(conn);
    return;
  }

  conn->cliwr.buf = conn->clird.buf;
  conn->cliwr.len = conn->clird.res;
  nb_SendWriteReq(&conn->cliwr);
}

static void wroteContent(NB_IOMsg * wmsg)
{
  Connection * conn = F2S(wmsg, clird, Connection);
  
  if (wmsg->res != wmsg->len)
  {
    localHttpReplyX(conn, "RM_XWRITEFAILED");
    return;
  }

  conn->conlen -= wmsg->res;

  if (conn->conlen)
    nb_SendReadReq(&conn->clird);
  else /* start reading data from server */
    nb_SendReadReq(&conn->serrd);
}

/*
 * Quite similar code than in s/ser/cli/ -functions 
 */

static int serHeader(Connection * conn, char * p);

static void initialSerInput(NB_IOMsg * rmsg)
{
  Connection * conn = F2S(rmsg, serrd, Connection);
  int len = rmsg->res;
  char * p = rmsg->buf;

  TRACEF(TRCF_FUNCENTRY, ("initialSerInput(imsg = %p)\n", rmsg));

  if (len <= 0)
  {
    deleteConnection(conn);
    return;
  }

#ifdef DEBUG
  p[len] = '\0';
#endif

  while (len--)
  {
    if (*p++ == '\n')
    {
      rmsg->res = len;
      if (serHeader(conn, p))
	return;
    }
  }

  /*
   * Input buffer ended. Read more to it if still here (reading more headers).
   */

  if (p >= conn->bufstart + DBUFSIZE - 1)
  {
    userlog(ULOG_ERR,  "%s had more that %d bytes of header information\n",
	    "(update)", (DBUFSIZE - 1) - CHOFFSET);
    localHttpReply(conn, RM_TOOLONGHEADER);
    return;
  }

  rmsg->buf = p;
  rmsg->len = (DBUFSIZE - 1) - (p - conn->bufstart);

  nb_SendReadReq(rmsg);
}

static void serHeaderWritten(NB_IOMsg * wmsg);

static int serHeader(Connection * conn, char * p)
{
  unsigned char * buf = (unsigned char *)conn->serrd.buf;
  int len = (unsigned char *)p - buf;
  int i;

  TRACEF(TRCF_FUNCENTRY, ("serHeader(conn = %p, p = %p)\n", conn, p));

  conn->serrd.buf = p; /* next line */

  if (iscntrl(buf[0])) { /* check EOF case, otherwise error */
    if ((buf[0] == '\r' && buf[1] == '\n') || buf[0] == '\n')
    {
      ((char *)conn->serwr.buf)[0] = '\r';
      ((char *)conn->serwr.buf)[1] = '\n';

      conn->serwr.fd = conn->clird.fd;
      conn->serwr.len = (char *)conn->serwr.buf + 2 - conn->bufstart;
      conn->serwr.buf = conn->bufstart;
      conn->serwr.func = serHeaderWritten;

#ifdef SHOWHDRS
    WriteCS(1, "\n*** Modified Server Headers follows..."); PRTCONNID();
    write(1, conn->bufstart, conn->serwr.len);
#endif      
      nb_SendWriteReq(&conn->serwr);
      return TRUE; /* eoh */
    }
    /* else */
    localHttpReply(conn, RM_TRASHEDHEADER);
    return TRUE;
  }

  i = strmapCmp(serhdrmaptable, serhdrmaptablelen, buf);

#ifdef SHOWHDRS
  if (i == SM_HTTP) {
    WriteCS(1, "\n*** Original Server Headers follows..."); PRTCONNID();
  }
  write(1, buf, len);
#if 0  /* some debugging info */
  printf("input index: %d, str: %s\n", i, serhdrtable[i].string);
#endif
#endif

  /* here headers that are first copied as is. at the second
     switch below, there might be a few changes. Note that
     the entries below changes the value of buf and conn->serwr.buf,
     but others (for example SM_LOC and SM_URI) at the second switch
     operates w/ original buf and conn->serwr.buf -values */
  switch (i)
  {
  case SM_HTTP:
  case SM_SERVER:
  case SM_MIMEVER:
  case SM_CONTYPE:
  case SM_CONLEN:
    memcpy(conn->serwr.buf, buf, len);
    buf = conn->serwr.buf;
    conn->serwr.buf = buf + len;
    break;
  }

  switch (i)
  {
  case SM_HTTP:
    buf += serhdrlen(i);
    /* change possible HTTP/xxx header to HTTP/1.0 header */
    if (!isspace(buf[0]) && !isspace(buf[1]) && !isspace(buf[2]))
    { buf[0] = '1'; buf[1] = '.'; buf[2] = '0'; }
    buf += 3;
    while (!isspace(*buf))
      *buf++ = ' ';
    break;

  case SM_LOC:
  case SM_URI:
    i = serhdrlen(i);

    while (isspace(buf[i]))
      i++;
    if (buf[i] == '<') /* URI header has this... */
      i++;
    while (isspace(buf[i]))
      i++;

    memcpy(conn->serwr.buf, buf, i);
    conn->serwr.buf += i;
    buf += i;
    len -= i;

    /*
     * The code here handles server redirection messages only partially
     * that is, when the redirection destination begins with `http://'
     * In a (very rare) case where, for example, a hand-given URL is
     * redirected by server without beginning http:// the actual destination
     * host (and port) cannot be determined (since there is no Referer: header)
     */
    if (strncasecmp(buf, "http://", 7) == 0)
    {
      *(char *)conn->serwr.buf++ = '/';
      if (conn->sh_referer)
	*(char *)conn->serwr.buf++ = conn->sh_referer;
      *(char *)conn->serwr.buf++ = ':';

      buf += 7;
      len -= 7;
    }
    memcpy(conn->serwr.buf, buf, len);
    conn->serwr.buf += len;

    break;

  case SM_CONLEN:
    conn->conlen = atoi(buf + serhdrlen(i));
    /* perhaps the value above should be checked... */
    break;

  case SM_CONN:
    CSTPCPY(conn->serwr.buf, "Connection: close\r\n");
    break;

  case SM_CONTYPE:
    buf += serhdrlen(i);
    while (isspace(*buf))
      buf++;
    if (strncmp(buf, "text/html", 9) == 0)
      conn->sh_ishtml = 1;
    break;
  }
  return FALSE;
}

static void serContent(NB_IOMsg * rmsg);
static void serContentAsis(NB_IOMsg * rmsg);
static void serContentWritten(NB_IOMsg * wmsg);
static void serContentWrittenAsis(NB_IOMsg * wmsg);

/*          wroteSerHeader() ? */
static void serHeaderWritten(NB_IOMsg * wmsg)
{
  Connection * conn = F2S(wmsg, serwr, Connection);
  
  TRACEF(TRCF_FUNCENTRY, ("serHeaderWritten(wmsg == %p)\n", wmsg));

  if (wmsg->res != wmsg->len)
  {
    /* userlog? */
    deleteConnection(conn);
    return;
  }

  conn->serwr.func = serContentWritten;

  if (conn->sh_ishtml)
  {
    conn->serrd.func = serContent;
    conn->serwr.func = serContentWritten;
  }
  else
  {
    conn->serrd.func = serContentAsis;
    conn->serwr.func = serContentWrittenAsis;
  }

  conn->sb_state = 0;
  conn->sb_bufp = NULL;
  
  if (conn->serrd.res)
  {
/*    nb_SendReplyMsg((NB_Msg *)&conn->serrd); */
    conn->serrd.func(&conn->serrd);
  }
  else
  {
    conn->serrd.buf = conn->bufstart;
    conn->serrd.len = DBUFSIZE;
    
    nb_SendReadReq(&conn->serrd);
  }
}



static void handleTag(char * p)
{
  TRACEF(TRCF_FUNCENTRY, ("handleTag(p = %p)\n", p));
  
  if (*p == '/') /* netscape allows </   H1 > etc */
    for (p++; isspace(*p); p++)
      ;

  /**/ if (strncasecmp(p, "script", 6) == 0)	CSTRCPY(p, "tpircs");
  else if (strncasecmp(p, "applet", 6) == 0)	CSTRCPY(p, "telppa");
  else if (strncasecmp(p, "embed",  5) == 0)	CSTRCPY(p, "debme");
}      

static void handleContent(char * p, int len)
{
  while (len)
  {
    while (!isalpha(*p))
    {
      p++; len--;
      if (!len)
	return;
    }
    
    if (*p == 'h' || *p == 'H')
    {
      if (strncasecmp(p, "http://", 7) == 0)
      {
	p[0] = '/'; p[1] = 't'; p[2] = 't'; p[3] = 'p';
	p+= 7; len -= 7;
      }
    }
    else if ((*p == 'o' || *p == 'O') && (p[1] == 'n' || p[1] == 'N'))
      do
      {
	int l;
	char * q = p + 2;
      
	/**/ if (strncasecmp(q, "mouseover", 9) == 0) l = 11;
	else if (strncasecmp(q, "mouseout", 8) == 0) l = 10;
	else if (strncasecmp(q, "click", 5) == 0) l = 7;
	else if (strncasecmp(q, "change", 6) == 0) l = 8;
	else if (strncasecmp(q, "submit", 6) == 0) l = 8;
	else
	{ p+= 2; len-= 2; break; }
      
	p[0] = 'n'; p[1] = 'o';
	p+= l; len-= l;
      } while (0);

    while (len && isalpha(*p))
    {
      p++; len--;
    }
  }
}

static int oddNumberOfBackslashes(char * p)
{
  int i;

  for (i = 0; *p == '\\'; i++, p--)
    ;

  TRACEF(TRCF_FUNCEXIT, ("return %d = oddNumberOfBackslashes()\n", i & 1));
  
  return i & 1;
}

/*
 * Some of the code here is copied from Wget.
 */

/*
 * This is perhaps the most complex function (currently) in this
 * software. I'll comment more for the variables used there later.
 * Now I just want to get this program working...
 */

static void serContent(NB_IOMsg * rmsg)
{
  Connection * conn = F2S(rmsg, serrd, Connection);
  
  char * buf = rmsg->buf;
  int len = rmsg->res;
  int state = conn->sb_state;
  char * p = conn->sb_bufp;
  char c;
  
  TRACEF(TRCF_FUNCENTRY, ("serContent(rmsg = %p)\n", rmsg));

  if (len <= 0)
  {
    deleteConnection(conn);
    return;
  }

  do 
  {
    /* `continue' and `break' are synonyms below, but used for readability? */
    /* when `continue' is used, the control usually returns to switch */
    /* when `break' is used, do-while will always end (len == 0) */

    switch (state)
    {
    case 0: /* look for a tag */
      for (; len && *buf != '<'; buf++, len--);
      if (!len)	break;
      
      buf++; len--;
      state = 1;

    case 1: /* found (possibly) a starting tag */
      /* skip spaces */
      for (; len && isspace(*buf); buf++, len--);
      if (!len)	break;

      p = buf;
      buf++; len--;
      state = 2;

    case 2:
      if (*p == '/')
      {
	/* skip spaces... again */
	for (; len && isspace(*buf); buf++, len--);
	if (!len) break;
      }
      state = 3;

    case 3:
      /* find the tag end */
      for (; len && !isspace(*buf) && *buf != '>' && *buf != '=';
	   buf++, len--);
      if (!len)	break;

      if (*buf == '=')
      {
	/* <tag=something> is illegal.  Just skip it.  */
	buf++, len--;
	p = NULL;
	state = 0;
	continue;
      }
      if (p == buf)
      {
	/* *buf == '>'.  */
	buf++, len--;
	p = NULL;
	state = 0;
	continue;
      }

      handleTag(p);

      c = *buf; buf++; len--;
      
      if (c == '>')
      {
	p = NULL;
	state = 0;
	continue;
      }

      state = 4;

    case 4:
      /* skip spaces. */
      for (; len && isspace(*buf); buf++, len--);
      if (!len)	break;

      p = buf;
      state = 5;

    case 5: /* printf("5\n"); */
      /* now we check for end of tag and then parse the contents...
       for the few strings we are interestred of */

      for (; len && *buf != '>' && *buf != '\"' && *buf != '\''; buf++, len--);
      if (!len) break;

      c = *buf;
      
      if (c == '>')
      {
	handleContent(p, buf - p);
	p = NULL;
	state = 0;
	continue;
      }

      buf++; len--;
      
      if (c == '\"')
      {
	state = 7;
	goto case_7; /* see file GOTO_AVOIDANCE */
      }

      /* else *buf == '\'' */
      state = 6;

    case 6: /* printf("6\n"); */
      for (; len && *buf != '\'' && *buf != '\n'; buf++, len--);
      goto case_7b; /* see file GOTO_AVOIDANCE */

    case 7: /* printf("7\n"); */
    case_7:
      for (; len && *buf != '\"' && *buf != '\n'; buf++, len--);

    case_7b:
      if (!len) break;

      c = *buf; buf++; len--;

      if (c != '\n' && oddNumberOfBackslashes(buf - 2))
	continue;

      state = 5;
      continue;
    }
  } while (len);

  conn->sb_state = state;
  conn->sb_bufp = p;

  /* write either the whole buffer (to client) or fully handled part; */
  conn->serwr.buf = conn->serrd.buf;

  if (p == NULL)
    len = conn->serrd.res;
  else
  {
    /*
     * when here, pointer value of p is in range
     * conn->bufstart <-> conn->bufstart + DBUFSIZE -1, and buf is in range
     * conn->bufstart <-> conn->bufstart + DBUFSIZE, the only way the test
     * below may success is that p == conn->bufstart and
     * buf == conn->bufstart + DBUFSIZE (unless buggy software, someone
     * care to `purify'...)
     */
    if (buf == p + DBUFSIZE) /* inside tag, which does not fit into buffer */
    { /* tag too long. Break possible attribute and flush the data */
      buf[-1] = '.';
      len = conn->serrd.res; /* == serrd.len */
      /* no reset for state, but reset ptrs so that full buffer can be read */
      /* operations in function serContentWritten() are closely bound to this*/
      /* conn->ser.bufp = p; */
      conn->serrd.buf = p;
      handleContent(p, DBUFSIZE);
    }
    else
    {
      len = p - (char *)conn->serrd.buf;
      conn->serrd.buf = buf;
      
      if (len == 0)
      {
	serContentWritten(&conn->serwr); /* could use nb_ReplyMsg here ... */
	return;			/* ... also in similar places elsewhere */
      }
    }
  }
  conn->serwr.len = len;
    
  nb_SendWriteReq(&conn->serwr);
}
      
      
static void serContentWritten(NB_IOMsg * wmsg)
{
  Connection * conn = F2S(wmsg, serwr, Connection);
  char * p = conn->sb_bufp;
  
  if (wmsg->res != wmsg->len)
  {
    /* userlog? */
    deleteConnection(conn);
    return;
  }

  if (p == NULL)
  {
    conn->serrd.buf = conn->bufstart;
    conn->serrd.len = DBUFSIZE;
  }
  else
  {
    int space = conn->bufstart + DBUFSIZE - (char *)conn->serrd.buf;
    
    if (space > DBUFSIZE / 4 || p == conn->bufstart)
      conn->serrd.len = space;
    else
    {
      int len = (char *)conn->serrd.buf - p;
      int mov = p - conn->bufstart;

      memmove(conn->bufstart, p, len);

      conn->sb_bufp = conn->bufstart;
      conn->serrd.buf-= mov;
      conn->serrd.len+= mov;
    }
  }
  nb_SendReadReq(&conn->serrd);
}


static void serContentAsis(NB_IOMsg * rmsg)
{
  Connection * conn = F2S(rmsg, serrd, Connection);

  TRACEF(TRCF_FUNCENTRY, ("serContentAsis(rmsg = %p)\n", rmsg));

  if (rmsg->res <= 0)
  {
    /* userlog? */
    deleteConnection(conn);
    return;
  }

  conn->serwr.buf = rmsg->buf;
  conn->serwr.len = rmsg->res;
  
  nb_SendWriteReq(&conn->serwr);
}

static void serContentWrittenAsis(NB_IOMsg * wmsg)
{
  Connection * conn = F2S(wmsg, serwr, Connection);
  
  TRACEF(TRCF_FUNCENTRY, ("serContentWrittenAsis(rmsg = %p)\n", wmsg));

  if (wmsg->res != wmsg->len)
  {
    /* userlog? */
    deleteConnection(conn);
    return;
  }

  nb_SendReadReq(&conn->serrd);
}
