#include "project.h"

RCID("$Id: netsupport.c $");
/*
 * Author: Tomi Ollila <too@iki.fi>
 *
 * Created: Mon Mar 23 16:37:15 1998 too
 * Last modified: Mon Sep  7 19:47:42 1998 too
 *
 * HISTORY
 * $Log: $
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <ctype.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>

#include "nbcore.h"
#include "prgutil.h"
#include "userlog.h"
#include "debugtools.h"

/* this uses blocking calls ... */
struct in_addr inetAddress(char * arg)
{
  struct in_addr	addr;
  struct hostent *	remote;

  TRACEF(TRCF_FUNCENTRY, ("inetAddress(addr = %s)\n", arg));

  if ((remote = gethostbyname(arg)) != NULL)
    memcpy(&addr, remote->h_addr, sizeof addr);
  else if ((addr.s_addr = inet_addr(arg)) == (unsigned long)-1)
    userlog(4, "Cannot resolve host address `%s'\n", arg);

  return addr;
}



static int ONE = 1;

/* should that nb_SetFd() be done elsewhere ? */

int inetSocket()
{
  int sd = socket(AF_INET, SOCK_STREAM, 0);

  TRACEF(TRCF_FUNCENTRY, ("inetSocket()\n"));

  if (sd >= 0)
    nb_SetFd(sd);
  else
    userlog(0, "socket() failed: %m\n");

  return sd;
}

int inetBind(int port)
{
  int fd;
  struct sockaddr_in addr = { 0 };

#if 1 /* SECURITY FEATURE IN THIS BETA RELEASE !!! */ 
  addr.sin_addr.s_addr = inet_addr("127.0.0.1");
#endif  
  
  TRACEF(TRCF_FUNCENTRY, ("inetBind(port = %d)\n", port));

  if ((fd = inetSocket()) < 0)
    Exit();

  setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &ONE, sizeof ONE);

  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);

  if (bind(fd, &addr, sizeof addr) < 0)
    pfatal("bind");

  listen(fd, 5);

  return fd;
}


int inetAccept(int fd)
{
#if 0
  struct sockaddr_in addr = { 0 };
  int len = sizeof addr;
  int sd = accept(fd, &addr, &len);
#else
  int sd = accept(fd, NULL, NULL); /* not interested at this time */
#endif

  TRACEF(TRCF_FUNCENTRY, ("inetAccept(fd = %d)\n", fd));

  if (sd < 0)
  {
    userlog(0, "accept failed: %m\n");
    return -1;
  }

#if 0
  TRACEL(5, ("Accepted from %s, ", inet_ntoa(addr.sin_addr)));
#endif

  nb_SetFd(sd);

  return sd;
}

int inetConnect(int fd, struct in_addr host, int port)
{
  struct sockaddr_in addr;

  TRACEF(TRCF_FUNCENTRY, ("inetConnect(fd = %d, host = %s, port = %d)\n",
			  fd, inet_ntoa(addr.sin_addr), port));


  addr.sin_family = AF_INET;
  addr.sin_addr = host;
  addr.sin_port = htons(port);

  nb_SetFd(fd);

  setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &ONE, sizeof ONE);

  if (connect(fd, &addr, sizeof addr) >= 0)
    return 1;

  TRACEL(5, ("errno == %d\n", errno));
  if (errno != EINPROGRESS)
  {
    userlog(0, "connect() failed: %m\n");
    return -1;
  }

  return 0; /* connection pending */
}

char * writePeerName(char * buf, int fd)
{
  struct sockaddr_in addr;
  int addrlen = sizeof addr;

  if (getpeername(fd, &addr, &addrlen) < 0)
    return NULL;

  strcpy(buf, inet_ntoa(addr.sin_addr));

  return buf + strlen(buf);
}
