#include "project.h"

RCID("$Id: localreplies.c $");
/*
 * Author: Tomi Ollila <too@iki.fi>
 *
 * Created: Fri Aug  7 18:18:23 1998 too
 * Last modified: Thu Sep 10 19:37:25 1998 too
 *
 * HISTORY
 * $Log: $
 */

#include <string.h>
#include <ctype.h>
#include <sys/utsname.h>

#include "localreplies.h"
#include "clihdrtables.h"
#include "strstuff.h"

char * writeHtmlStart(char * buf, char * title, int titlen)
{
  CSTPCPY(buf, "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML//EN\">\n"
	  "<html><head>\n<title>");

  memcpy(buf, title, titlen);
  buf += titlen;

  CSTPCPY(buf, "</title>\n</head><body>\n<h1>");

#if 0  /* vvvvv well, perhaps not needed vvvvv */
  /* if title has xxx reply code, remove it for page header */
  if (isdigit(title[0]))
  {
    title += 4;
    titlen -= 4;
  }
#endif

  memcpy(buf, title, titlen);
  buf += titlen;

  CSTPCPY(buf, "</h1>\n\n");

  return buf;
}

char * writeUserAgent(char * info, Connection * conn, char * buf)
{
  char * ua = findHeader(conn->bufstart + CHOFFSET, CM_USERAGENT);
  int len;

  if (!ua)
    return buf;

  buf = stpcpy(buf, info);

  *buf++ = '<';
  *buf++ = 'I';
  *buf++ = '>';

  len = (ua[1] << 8) + ua[2];

  ua += clihdrtable[CM_USERAGENT].strlen;
  len -= clihdrtable[CM_USERAGENT].strlen;

  do
    len--;
  while (isspace(ua[len]));
  len++;

  memcpy(buf, ua, len);
  buf += len;

  CSTPCPY(buf, ".</I>\n<P>\n");

  return buf;
}

/*
 * It might be smart to evaluate this function only once at startup
 * and then just set up global pointers for it (P.sysinfo and P.sysinfolen)
 */

char * writeSystemInfo(Connection * conn, char * buf)
{
  struct utsname u;

  if (uname(&u) < 0)
    return buf;

  CSTPCPY(buf, "\n<HR>\n<DIV ALIGN=right>\n<B>" VERSION "</B> ");

  buf = stpcpy(buf, u.nodename);

  *buf++ = ' ';  *buf++ = '(';

  buf = stpcpy(buf, u.sysname);

  *buf++ = ' ';

  buf = stpcpy(buf, u.release);

  *buf++ = ',';  *buf++ = ' ';

  buf = stpcpy(buf, u.machine);

  CSTPCPY(buf, ")\n</DIV>");

  return buf;
}


extern struct LanguageFunctions languagefunctions_en;
extern struct LanguageFunctions languagefunctions_fi;

/* add pointers to your language-entries here */
static struct {
  char * lname;
  struct LanguageFunctions * lfuncs;
} languagefunctiontable[] = {
  { "fi", &languagefunctions_fi },
  { NULL, &languagefunctions_en } /* fallback */
};

static void reply_and_delete(Connection * conn, const char * buf, int len);

void localHttpReplySysErr(Connection * conn, char * f, int errno)
{
  /* reusing fields... should use other things but this is a fast hack now */
  /* actually, these variables could even be put into GLOBAL VARIABLE */
  
  ERRVAL = errno;
  ERRSTR = f;
  localHttpReply(conn, RM_SYSERR);
}

void localHttpReplyX(Connection * conn, char * xcode)
{
  ERRSTR = xcode;
  localHttpReply(conn, RM_NOMSG);
}

void localHttpReply(Connection * conn, int mi)
{
  char * buf = conn->bufstart;
  struct LanguageFunctions * lf;
  int i;
  char * alang = findHeader(conn->bufstart + CHOFFSET, CM_ACCLANG);
  char * title;
  int titlen;
  
  TRACEF(TRCF_FUNCENTRY, ("localHttpReply(conn = %p, mi = %d)\n", conn, mi));

  /* check if there is support for requested language ...*/
  if (alang == NULL)
    alang = "";
  else {
    char * p = alang + clihdrtable[CM_ACCLANG].strlen;

    alang = p;
    while (isalpha(*p)) p++;
    *p = '\0'; 			/* ... up to major part (en-GB -> en etc.) */

/*    printf("**** alang!!!: %s ***\n", alang); */
  }

  for (i = 0; languagefunctiontable[i].lname; i++)
    if (strcmp(alang, languagefunctiontable[i].lname) == 0)
      break;

  lf = languagefunctiontable[i].lfuncs;

  CSTPCPY(buf, "HTTP/1.0 ");

  title = lf->getTitleH1(mi, &titlen);
  buf = httpReplyCode(buf, title);

  CSTPCPY(buf, "\r\n"
	  "Server: " VERSION "\r\n"
	  "Connection: close\r\n"
	  "Content-Type: text/html\r\n"
	  "\r\n");

#ifdef SHOWHDRS
  WriteCS(1, "\nLocal HTTP Reply Page Headers follows..."); PRTCONNID();
  write (1, conn->bufstart, buf - conn->bufstart);
#endif  

  buf = writeHtmlStart(buf, title, titlen);
  buf = lf->addBody(mi, conn, buf);
  buf = writeSystemInfo(conn, buf);

  CSTPCPY(buf, "\n</body>\n</html>\n");

  reply_and_delete(conn, conn->bufstart, buf - conn->bufstart);
}


#define NOHTTP1XTITLE "At least HTTP 1.0 compliant browser required"
void noHttp1x(Connection * conn)
{
  char * buf = conn->bufstart;

  buf = writeHtmlStart(buf, NOHTTP1XTITLE, sizeof NOHTTP1XTITLE -1);

  CSTPCPY(buf, "For proper operation, HTTPGate requires that"
	  "at least HTTP 1.0 capable browser is used\n"
	  "</body></html>\n");

  reply_and_delete(conn, conn->bufstart, buf - conn->bufstart);
}

#ifdef DEBUG
void showDataOnPage(Connection * conn, char * data, int dlen)
{
  char * buf = conn->bufstart;

  buf = writeHtmlStart(buf, "Show Data", 9);

  memcpy(buf, data, dlen);

  reply_and_delete(conn, conn->bufstart, buf + dlen - conn->bufstart);
}
#endif

char * redirect_start(Connection * conn)
{
  char * buf = conn->bufstart;

  CSTPCPY(buf, "HTTP/1.0 302 Redirection ");
  CSTPCPY(buf, "\r\n"
	  "Server: " VERSION "\r\n"
	  "MIME-version: 1.0\r\n"
	  "Connection: close\r\n"
	  "Location: ");

  return buf;
}

void redirect(Connection * conn, char * url, char * buf)
{
  int len = buf - url;

  TRACEF(TRCF_FUNCENTRY, ("redirect(conn = %p, url = %p, buf = %p)\n",
			  conn, url, buf));

  if (len > DBUFSIZE / 4)
  {
    localHttpReply(conn, RM_TOOLONGURL);
    return;
  }

  CSTPCPY(buf, "\r\nURI: <");

  memcpy(buf, url, len);
  buf+= len;

  CSTPCPY(buf, ">\r\nContent-type: text/html\r\n\r\n");

#ifdef SHOWHDRS
  WriteCS(1, "\nLocal Redirection Page Headers follows..."); PRTCONNID();
  write (1, conn->bufstart, buf - conn->bufstart);
#endif  
  
  CSTPCPY(buf, "<HEAD><TITLE>HTTPGate Redirection</TITLE></HEAD>\n"
	  "<BODY>\n<H1>HTTPGate Redirection</H1>\n"
	  "If you see this page, click\n<A HREF=\"");

  memcpy(buf, url, len);
  buf+= len;

  CSTPCPY(buf, "\">here</A>\nto get your requested page.\n"
	  "<P>\n</BODY>\n<HTML>\n");

  reply_and_delete(conn, conn->bufstart, buf - conn->bufstart);
}

static void deleteConnFromSerWR(NB_IOMsg * msg)
{
  deleteConnection(F2S(msg, serwr, Connection));
}


static void reply_and_delete(Connection * conn, const char * buf, int len)
{
  conn->serwr.fd = conn->clird.fd;
  conn->serwr.buf = (char *)buf; /* XXX const to non-const... */
  conn->serwr.len = len;
  conn->serwr.func = deleteConnFromSerWR;

  nb_SendWriteReq(&conn->serwr);
}
