/*
 * $Id: listcode.c $
 *
 * Author: Tomi Ollila <Tomi.Ollila@iki.fi>
 *
 * Created: Tue Mar 24 16:45:16 1998 too
 * Last modified: Tue Mar 24 16:51:21 1998 too
 *
 * HISTORY 
 * $Log: $
 */

/* used only when debugging (compile filan source w/ optimize flag on) */

#ifndef __OPTIMIZE__

#include "list.h"

void addHead (struct List * list, struct Node * node)
{
  list->lh_Head->ln_Pred = node;
  node->ln_Succ = list->lh_Head;
  node->ln_Pred = (struct Node *)&list->lh_Head;
  list->lh_Head = node;
}

void addTail (struct List * list, struct Node * node)
{
  list->lh_TailPred->ln_Succ = node;
  node->ln_Pred = list->lh_TailPred;
  node->ln_Succ = (struct Node *)&list->lh_Tail;
  list->lh_TailPred = node;
}

void insNodeP(struct Node * next, struct Node * node)
{
  struct Node * prev = next->ln_Pred;

  node->ln_Pred = prev;
  node->ln_Succ = next;
  prev->ln_Succ = node;
  next->ln_Pred = node;
}

void insNodeS(struct Node * prev, struct Node * node)
{
  struct Node * next = prev->ln_Succ;

  node->ln_Succ = next;
  node->ln_Pred = prev;
  prev->ln_Succ = node;
  next->ln_Pred = node;
}

void newList (struct List * list)
{
  list->lh_Head = (struct Node *)&list->lh_Tail;
  list->lh_Tail = NULL;
  list->lh_TailPred = (struct Node *)&list->lh_Head;
}

struct Node * remHead (struct List * list)
{
  struct Node * node = list->lh_Head;
  if (node->ln_Succ == NULL)
    return NULL;

  list->lh_Head = node->ln_Succ;
  node->ln_Succ->ln_Pred = (struct Node *)&list->lh_Head;

  return node;
}

struct Node * remHeadQ (struct List * list)
{
  struct Node * node = list->lh_Head;
  
  list->lh_Head = node->ln_Succ;
  node->ln_Succ->ln_Pred = (struct Node *)&list->lh_Head;

  return node;
}

void remNode (struct Node * node)
{
  struct Node *node2;

  node2 = node->ln_Succ;
  node = node->ln_Pred;
  node->ln_Succ = node2;
  node2->ln_Pred = node;
}

struct Node * remTail (struct List * list)
{
  struct Node * node = list->lh_TailPred;

  if (node == (struct Node *)list)
    return NULL;
  
  list->lh_TailPred = node->ln_Pred;
  node->ln_Pred->ln_Succ = (struct Node *)&list->lh_TailPred;

  return node;
}

struct Node * remTailQ (struct List * list)
{
  struct Node * node = list->lh_TailPred;

  list->lh_TailPred = node->ln_Pred;
  node->ln_Pred->ln_Succ = (struct Node *)&list->lh_TailPred;

  return node;
}

void swapLists(struct List * list1, struct List * list2)
{
  struct List tmplist = *list1;
  int empty1 = emptyList(list1);

  if (emptyList(list2))
    newList(list1);
  else {
    *list1 = *list2;
    list1->lh_Head->ln_Pred = (struct Node *)list1;
    list1->lh_TailPred->ln_Succ = (struct Node *)&list1->lh_Tail;
  }

  if (empty1)
    newList(list2);
  else {
    *list2 = tmplist;
    list2->lh_Head->ln_Pred = (struct Node *)list2;
    list2->lh_TailPred->ln_Succ = (struct Node *)&list2->lh_Tail;
  }
}

#endif /* __OPTIMIZE__ */

