/*
 * $Id: debugtools.c $
 *
 * Author: Tomi Ollila <Tomi.Ollila@iki.fi>
 *
 * Created: Tue Jan 17 19:32:48 1995 too
 * Last modified: Mon Apr  6 18:15:44 1998 too
 *
 * HISTORY 
 * $Log: $
 */

#ifdef DEBUG

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <time.h>

#include "userlog.h"
#include "prgutil.h"
#include "debugtools.h"

#if 1
long trcflags = 0;	/* none by default */
#else
long trcflags = -1;	/* all by default (also levels) */
#endif

#if 1
long trclevel = 0;	/* none by default */
#else
long trclevel = 0xff;	/* all by default */
#endif

void debugf(char * format, ...)
{
  va_list ap;
  
  va_start(ap, format);
  (void)vuserlog(7, format, (char *)ap);
  va_end(ap);
}

/* add `new global' to the beginning, and more like local from the end */
static const char * trcftable[] = {
  "funcentry",
  "funcexit",
  "scanlist",
  "select",
   "4",  "5",  "6",  "7",
   "8",  "9", "10", "11", "12", "13", "14", "15",
  "16", "17", "18", "19", "20", "21", "22", "23" };


static long parse_dbflags(char * fs)
{
  long flags = 0;
  
  while (*fs)
  {
    int addf;
    char fbuf[64];
    int flen;
    char * p;

    if (*fs == ',') fs++;
    
    if (*fs == '-')	{ addf = 0; fs++; }
    else		  addf = 1;

    if ((p = strchr(fs, ',')) != NULL)
      flen = p - fs;
    else
      flen = strlen(fs);

    if (flen < sizeof fbuf)
    {	
      int index;

      memcpy(fbuf, fs, flen);
      fbuf[flen] = '\0';
      fs += flen;
      
      if (memcmp(fbuf, "all", flen) == 0)
      { flags = 0xffffff00; continue; }
      
      if (memcmp(fbuf, "none", flen) == 0)
      { flags = 0x00000000; continue; }

      if ((index = arrstrstr(trcftable, fbuf)) >= 0)
      {
	int flag = 1 << (index + 8);

	if (addf) flags |= flag;
	else	  flags &= ~flag;
	continue;
      }
      userlog(5, "%s: unknown debug flag\n", fbuf);
      
      continue;
    }
    userlog(5, "mysterious debug flag information: %s\n", fs);
    fs += flen;
  }

  return flags;
}
      

static const char * argtable[] = { "-debug-flags", "-debug-level", NULL };
enum { ARG_DBFLAGS, ARG_DBLEVEL };

static int argfunc(PRG_Args * args, int n, int * argcp, char *** argvp)
{
  switch (n) {
  case ARG_DBFLAGS:
    trcflags = parse_dbflags(need_next_arg(argcp, argvp));
    break;
  case ARG_DBLEVEL:
    trclevel = atoi(need_next_arg(argcp, argvp));
    break;
  default: /* n == -2 or n == -1, do nothing */
  }
  return 0; /* success */
}

static PRG_Args dbargs = {
  NULL,
  argtable,
  argfunc
};

void debugArgs()
{
  prg_AddArgList(&dbargs);
}


#endif /* DEBUG */
