/*
 * $Id: debugtools.h $
 *
 * Author: Tomi Ollila <too@iki.fi>
 *
 * Created: Tue Jan 17 19:33:05 1995 too
 * Last modified: Sat Jul 11 13:47:14 1998 too
 *
 * HISTORY 
 * $Log: debugtools.h,v $
 */

#ifndef DEBUGTOOLS_H
#define DEBUGTOOLS_H

#ifdef DEBUG

extern long trcflags; /* flags from 0x100 to 0x80000000 */
extern long trclevel; /* levels from 1 to 255 */ /* value of 0 - no output */
extern void debugf(char * format, ...);

void debugArgs(void);

/* do - while here protects against if - else - else -problems */

#define assert(x) do if (!(x)) {\
  debugf("%s:%d: assertion failed: `%s'\n", __FILE__,__LINE__, #x); \
  abort(); } while (0)

#define TRACEF(x, y) do if (x & trcflags) \
{ debugf("%s:%d: ",__FILE__,__LINE__); debugf y; } while(0)

#define TRACEL(x, y) do if ((unsigned)(x-1) < trclevel) \
{ debugf("%s:%d: ",__FILE__,__LINE__); debugf y; } while(0)

#define TRACE(x, y) do if (((unsigned)(x-1) < trclevel) || (x & trcflags)) \
{ debugf("%s:%d: ",__FILE__,__LINE__); debugf y; } while(0)


/* flags, TRCFLAGS choses the group from below of which will generate output */
#define TRCF_NONE	0x0000
#define TRCF_FUNCENTRY	0x0100
#define TRCF_FUNCEXIT	0x0200
#define TRCF_SCANLIST	0x0400
#define TRCF_SELECT	0x0800

#define TRCF_N	0x0000
#define TRCF_0	0x00000100
#define TRCF_1	0x00000200
#define TRCF_2	0x00000400
#define TRCF_3	0x00000800
#define TRCF_4	0x00001000
#define TRCF_5	0x00002000
#define TRCF_6	0x00004000
#define TRCF_7	0x00008000
#define TRCF_8	0x00010000
#define TRCF_9	0x00020000
#define TRCF_10	0x00040000
#define TRCF_11	0x00080000

#else /* ! DEBUG */

#define TRACEF(x, y)
#define TRACEL(x, y)
#define TRACE(x, y)

#define assert(x)
#define debugArgs()

#endif /* DEBUG */

#endif /* DEBUGTOOLS_H */
