/*
 * $Id: nbcore.c,v 1.1 1996/06/29 17:10:55 too Exp too $
 *
 * Author: Tomi Ollila <too@iki.fi>
 *
 * Created: Wed Apr 24 20:23:32 1996 too
 * Last modified: Mon Sep  7 08:43:33 1998 too
 *
 * HISTORY 
 * $Log: nbcore.c,v $
 */

#ifdef DEBUG
#include "debugtools.h"
#else
#define TRACEF(x, y);
#define assert(x);
#endif

#define select(a,b,c,d,e) fooselect(a,b,c,d,e) /* HP-SuX */

#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>

#ifdef __sun__
#include <sys/filio.h>
#endif

#include "nbcore.h"

#if NB_IOMSGBUFSIZ
#define IOMSG	data
#else
#define IOMSG	buf
#endif


#undef select

const int ONE = 1;
const int ZERO = 0;

static struct
{
  int	 maxfd;
  fd_set fds;		/* bitfield of descriptors in use */
  fd_set readfds;	/* current read fd:s */
  fd_set writefds;	/* current write fd:s */
  void  (*sighandler)(int, void *);
  void * sigdata;
  char volatile	sigs[8];	/* vector of arrived signals */
} NBCPriv = { 0 };

struct NBCore NBCore = { 0 };

/*
 * this variable below have external scope.
 */

#if NB_HIRES_TIMERS
struct timeval NB_TIMEVAL;
#else
unsigned long  NB_TIME;
#endif  

#define remNodeZ(node) do { remNode(node); (node)->ln_Succ = NULL; } while (0)

/* Oh, we love HP-SuX */
extern int select(int, fd_set *, fd_set *, fd_set *, struct timeval *);
extern void memset(void *, int, int);

void nb_init()
{
#if 0
  memset(&NBCore, 0, sizeof NBCore);
  memset(&NBPriv, 0, sizeof NBPriv);
#endif  

  newList(&NBCore.addreadlist);
  newList(&NBCore.addwritelist);
  newList(&NBCore.timerlist);
  newList(&NBCore.replylist);

#if NB_HIRES_TIMERS
  gettimeofday(&NB_TIMEVAL, NULL);
#else    
  NB_TIME = time(NULL);
#endif
}

int nb_start()
{
  struct NB_Msg * msg;

#if NB_HIRES_TIMERS
  gettimeofday(&NB_TIMEVAL, NULL);
#else    
  NB_TIME = time(NULL);
#endif
  
  /* initialization functions may have added "level 2" initialization
   * hooks to nbcore reply list, which can be used in this purpose.
   * each "level 2" initialization functions may then add "level 3"
   * hooks, and so on. After no function adds more higher level
   * initialization hooks, the system is ready to go to main event loop
   */

  /* during this scan more items can be added to the list */
  while ((msg = (struct NB_Msg *)remHead(&NBCore.replylist))) {
    int i;
      
    msg->node.ln_Succ = NULL; /* mark not being in a list */
    if ((i = ((int (*)(NB_Msg *))(*msg->func))(msg)))
      return i; /* failure */
  }
  return 0;
}

static void nb_SigHandler(int s)
{
  int i;

  for (i = 0; i < 8; i++) {
    if (NBCPriv.sigs[i] == 0)      { NBCPriv.sigs[i] = s; return; }
    if (NBCPriv.sigs[i] == s)      return;
  }
}

void nb_main()
{
  List replylist;
  List readlist;
  List writelist;

  newList(&replylist);
  newList(&readlist); 
  newList(&writelist);

  FD_ZERO(&NBCPriv.readfds);
  FD_ZERO(&NBCPriv.writefds);

#if NB_HIRES_TIMERS
  gettimeofday(&NB_TIMEVAL, NULL);
#else    
  NB_TIME = time(NULL);
#endif
  
  while (1)
  {
    int n;
    fd_set   rfds, wfds;
    fd_set * wfdsp;
    struct timeval tv;
    struct NB_TimeMsg * tm;
    struct NB_Msg * msg;
    struct NB_IOMsg * rwmsg;

    /* these scans could be removed if using mark in lists and ... */
    while ((rwmsg = (NB_IOMsg *)remHead(&NBCore.addreadlist))) {
      addTail(&readlist, (Node *)rwmsg);
      FD_SET(rwmsg->fd, &NBCPriv.readfds);
    }
    rfds = NBCPriv.readfds;
    
    while ((rwmsg = (NB_IOMsg *)remHead(&NBCore.addwritelist))) {
      addTail(&writelist, (Node *)rwmsg);
      FD_SET(rwmsg->fd, &NBCPriv.writefds);
    }

    if (emptyList(&writelist))
      wfdsp = NULL;
    else {
      wfdsp = &wfds;
      wfds = NBCPriv.writefds;
    }

    tm = (struct NB_TimeMsg *)NBCore.timerlist.lh_Head;

    if (!emptyList(&NBCore.replylist)) {
      tv.tv_sec = tv.tv_usec = 1;
      TRACEF(TRCF_SELECT, ("select1: maxfd = %d, rfds = %lx\n",
			   NBCPriv.maxfd, *(int *)&rfds));
      n = select(NBCPriv.maxfd, &rfds, wfdsp, NULL, &tv);
      TRACEF(TRCF_SELECT, ("select1: n = %d, rfds = %lx\n", n, *(int *)&rfds));
    }
    else {
      if (tm->node.ln_Succ == NULL) {
	TRACEF(TRCF_SELECT, ("select2: maxfd = %d, rfds = %lx\n",
			     NBCPriv.maxfd, *(int *)&rfds));
	n = select(NBCPriv.maxfd, &rfds, wfdsp, NULL, NULL);
	TRACEF(TRCF_SELECT, ("select2: n = %d, rfds = %lx\n",
			     n, *(int *)&rfds));
      }
      else {
#define EXTRAWAIT 5000

#if NB_HIRES_TIMERS	
	TRACEF(TRCF_N, ("%d.%d -- %d.%d\n", tm->sec, tm->usec,
			NB_TIMEVAL.tv_sec, NB_TIMEVAL.tv_usec));

	tv.tv_sec = tm->sec - NB_TIMEVAL.tv_sec;
	tv.tv_usec = tm->usec - NB_TIMEVAL.tv_usec + EXTRAWAIT ;
	if (tv.tv_usec < 0) {
	  tv.tv_usec+= 1000000;
	  tv.tv_sec--;
	}
	else if (tv.tv_usec > 1000000) {
	  tv.tv_usec-= 1000000;
	  tv.tv_sec++;
	}

	if ((long)tv.tv_sec < 0)
	  tv.tv_sec = tv.tv_usec = 0;
#else
	TRACEF(TRCF_N, ("%d -- %d\n", tm->sec, NB_TIME));

	tv.tv_sec = tm->sec - NB_TIME;
	tv.tv_usec = EXTRAWAIT;
#endif
	
	TRACEF(TRCF_SELECT, ("select3  maxfd: %d, rfds: %lx, tv: %d:%d\n",
		    NBCPriv.maxfd, *(int *)&rfds, tv.tv_sec, tv.tv_usec));
	n = select(NBCPriv.maxfd, &rfds, wfdsp, NULL, &tv);
	TRACEF(TRCF_SELECT, ("select3  n: %d, rfds: %lx\n", n, *(int *)&rfds));
      }
    }

    while (*(int *)&NBCPriv.sigs)
    {
      int i;
      
      for (i = 0; i < 8; i++)
	if (NBCPriv.sigs[i]) {
	  NBCPriv.sighandler(NBCPriv.sigs[i], NBCPriv.sigdata);
	  NBCPriv.sigs[i] = 0;
	}
	else
	  break;
    }

    if (n < 0 && errno != EINTR)
      return;
    
#if NB_HIRES_TIMERS
    gettimeofday(&NB_TIMEVAL, NULL);
#else    
    NB_TIME = time(NULL);
#endif
    
    if (tm->node.ln_Succ != NULL) {
      Scan_List_BEGIN(&NBCore.timerlist, struct NB_TimeMsg *, tm, 2)
#if NB_HIRES_TIMERS	
	if ((NB_TIMEVAL.tv_sec == tm->sec && NB_TIMEVAL.tv_usec >= tm->usec) ||
	    NB_TIMEVAL.tv_sec > tm->sec)
#else	  
	if (NB_TIME >= tm->sec)
#endif
	{
	  remNodeZ((Node *)tm);
	  tm->func(tm);
	}
	else
	  break;
      Scan_List_END;
    }

    if (n > 0) {
      Scan_List_BEGIN(&readlist, struct NB_IOMsg *, rm, 2) {
	TRACEF(20, ("rm->fd = %d\n", rm->fd));
	if (FD_ISSET(rm->fd, &rfds)) {
	  /* with -O2, gcc can register allocate structure field variables */
	  TRACEF(20, ("rm->len = %d\n", rm->len));
	  if (rm->len > 0)
	    do
	      rm->res = read(rm->fd, rm->IOMSG, rm->len);
	    while (rm->res < 0 && errno == EINTR);
	  remNodeZ((Node *)rm);
	  FD_CLR(rm->fd, &NBCPriv.readfds);
	  rm->func(rm);
	}
      } Scan_List_END;

      if (wfdsp) {
	Scan_List_BEGIN(&writelist, struct NB_IOMsg *, wm, 2) {
	  TRACEF(20, ("wm->fd = %d\n", wm->fd));
	  if (FD_ISSET(wm->fd, &wfds)) {
	    int len = 0;
	    
	    TRACEF(20, ("wm->len = %d, wm->res = %d\n", wm->len, wm->res));
	    if (wm->len > 0) {
	      do
		len = write(wm->fd,
			    (char *)wm->IOMSG + wm->res, wm->len - wm->res);
	      while (len < 0 && errno == EINTR);
	      if (len <= 0)
		len = 0;
	      else
		wm->res += len;
	    }
	    /* len == 0 must be first below, (may be using NB_IOEventMsg *) */
	    if (len == 0 || wm->res == wm->len) {
	      remNodeZ((Node *)wm);
	      FD_CLR(wm->fd, &NBCPriv.writefds);
	      wm->func(wm);
	    }
	  }
	} Scan_List_END;
      }
    }

    /* here replylist is scanned only once during loop, so functions
       may add new items to the list */
    if ((msg = (struct NB_Msg *)remHead(&NBCore.replylist))) {
      swapLists(&NBCore.replylist, &replylist);
      do {
	msg->node.ln_Succ = NULL; /* mark not being in a list */
	msg->func(msg);
      } while ((msg = (struct NB_Msg *)remHead(&replylist)));
    }
  }
}

/***************************************************************************
 * Send requests
 */

#if NB_HIRES_TIMERS

void nb_SendTimeReq(struct NB_TimeMsg * tm)
{
  Scan_List_BEGIN(&NBCore.timerlist, struct NB_TimeMsg *, ltm, 0)
    if (tm->sec < ltm->sec ||
	(tm->sec == ltm->sec && tm->usec < ltm->usec)) {
      insNodeP((Node *)ltm, (Node *)tm);
      return;
    }
  Scan_List_END;
  addTail(&NBCore.timerlist, (Node *)tm);
}

void nb_SendTimeReq2(struct NB_TimeMsg * tm, unsigned long sec, long usec)
{
  long u;

  u = NB_TIMEVAL.tv_usec + usec;
  
  if (u > 1000000) {
    sec += 1;
    u -= 1000000;
  }
  tm->usec = u;
  tm->sec = NB_TIMEVAL.tv_sec + sec;

  nb_SendTimeReq(tm);
}

#else	/* not NB_HIRES_TIMERS */

void nb_SendTimeReq(struct NB_TimeMsg * tm)
{
  Scan_List_BEGIN(&NBCore.timerlist, struct NB_TimeMsg *, ltm, 0)
    if (tm->sec < ltm->sec) {
      insNodeP((Node *)ltm, (Node *)tm);
      return;
    }
  Scan_List_END;
  addTail(&NBCore.timerlist, (Node *)tm);
}

#endif	/* NB_HIRES_TIMERS */


/***************************************************************************
 * Abort requests
 */

void nb_AbortMsg(struct NB_Msg * msg)
{
  TRACEF(TRCF_FUNCENTRY, ("nb_AbortMsg(struct NB_Msg * msg = %lx)\n", msg));

  if (msg->node.ln_Succ) {
    remNode(&msg->node);
    msg->node.ln_Succ = NULL;
  }
}

void nb_AbortReadReq(struct NB_IOMsg * rm)
{
  TRACEF(TRCF_FUNCENTRY,("nb_AbortReadMsg(struct NB_IOMsg * rm = %lx)\n", rm));

  if (rm->node.ln_Succ) {
    TRACEF(20, ("AbortReadMsg: fd == %d\n", rm->fd));

    if (!FD_ISSET(rm->fd, &NBCPriv.readfds))
      TRACEF(0, ("Read msg fd %d without bit sen in readfds.\n", rm->fd));

    FD_CLR(rm->fd, &NBCPriv.readfds);
    remNode(&rm->node);
    rm->node.ln_Succ = NULL;
  }
}

void nb_AbortWriteReq(struct NB_IOMsg * wm)
{
  TRACEF(TRCF_FUNCENTRY,
	 ("nb_AbortWriteMsg(struct NB_IOMsg * wm = %lx)\n", wm));

  if (wm->node.ln_Succ) {
    TRACEF(20, ("AbortWriteMsg: fd == %d\n", wm->fd));

    if (!FD_ISSET(wm->fd, &NBCPriv.writefds))
      TRACEF(7, ("Write msg fd %d without bit sen in writefds.\n", wm->fd));

    FD_CLR(wm->fd, &NBCPriv.writefds);
    remNode(&wm->node);
    wm->node.ln_Succ = NULL;
  }
}

/***************************************************************************
 * Set operation parameters
 */

#include <stdarg.h>
void nb_SetSignals(int sig, ...) /* to end chain, set last as zero (0) */
{
  va_list ap;

  va_start(ap, sig);
  do
    signal(sig, nb_SigHandler);
  while ((sig = va_arg(ap, int)));
  va_end(ap);
}

void nb_SetSigHandler(void (*handler)(int, void *), void * data)
{
  NBCPriv.sighandler = handler;
  NBCPriv.sigdata = data;
}

#include <sys/ioctl.h>

void nb_SetFd(int fd)
{
  long arg;

  assert(fd >= 0);
  
  if (fd >= NBCPriv.maxfd)
    NBCPriv.maxfd = fd + 1;
  
  FD_SET(fd, &NBCPriv.fds);

#if 1  
  if (fcntl(fd, F_GETFL, &arg) >= 0)
    fcntl(fd, F_SETFL, arg | O_NONBLOCK); /* posix style works on most plats */
#else  
  ioctl(fd, FIONBIO, &ONE); /* set non-blocking IO */
#endif  
  ioctl(fd, FIOASYNC, &ZERO); /* clear asynchronous IO (if possible) */
}

/***************************************************************************
 * Close
 */

#ifdef close
#undef close /* check that there is no references to `close()' above */
#endif

/* Do not have files, which have more than one reference fd */
/* but this function may be called, even fd has not been set */
void nb_CloseFd(int fd)
{
  close(fd);

  if (!FD_ISSET(fd, &NBCPriv.fds)) {
    TRACEF(1, ("FYI: nb_CloseFd: Descriptor %d not in NB use set\n", fd));
    return;
  }
  
  FD_CLR(fd, &NBCPriv.fds);

  if (fd == NBCPriv.maxfd - 1) {
  /* if fd was for example 33, after it is closed, greatest possible is 32 */
    int i = howmany(fd - 1, NFDBITS) - 1;

    for (;i >= 0; i--) {
      fd_mask cmask = NBCPriv.fds.fds_bits[i];
      fd_mask rmask;

      if (!cmask) continue;

      fd = NFDBITS * i;
      if ((rmask = cmask & 0xFFFF0000)) { fd+= 16; cmask = rmask; }
      if ((rmask = cmask & 0xFF00FF00)) { fd+= 8;  cmask = rmask; }
      if ((rmask = cmask & 0xF0F0F0F0)) { fd+= 4;  cmask = rmask; }
      if ((rmask = cmask & 0xCCCCCCCC)) { fd+= 2;  cmask = rmask; }
      if (cmask & 0xAAAAAAAA) fd+= 1;

      NBCPriv.maxfd = fd + 1;
      return;
    }
    NBCPriv.maxfd = 0;
  }
}

#ifdef DEBUG
void nb_ChkClose(int fd)
{
  if (FD_ISSET(fd, &NBCPriv.fds)) {
    TRACEF(1, ("Descriptor %d closed directly\n", fd));
    nb_CloseFd(fd);
  }
  else
    close(fd);
}
#endif

/*************************************************************************/
/* There are inline euqivalents of the following functions */

#ifndef __OPTIMIZE__

#if NB_HIRES_TIMERS

void nb_SendTimeReq1(struct NB_TimeMsg * tm, unsigned long sec)
{
  tm->sec = NB_TIMEVAL.tv_sec + sec;
  nb_SendTimeReq(tm);
}

#else	/* not NB_HIRES_TIMERS */

void nb_SendTimeReq1(struct NB_TimeMsg * tm, unsigned long sec)
{
  tm->sec = NB_TIME + sec;
  nb_SendTimeReq(tm);
}

void nb_SendTimeReq2(struct NB_TimeMsg * tm, unsigned long sec, long usec)
{
  tm->sec = NB_TIME + sec;
  nb_SendTimeReq(tm);
}

#endif

/****/

void nb_SendReadReq(struct NB_IOMsg * rm)
{
  TRACEF(0, ("rm->len = %d\n", rm->len));
#ifdef DEBUG  
  if (!FD_ISSET(rm->fd, &NBCPriv.fds)) {
    TRACEF(1, ("Descriptor %d not set ready for nb_SendReadReq\n", rm->fd));
    nb_SetFd(rm->fd);
  }
#endif  
  addTail(&NBCore.addreadlist, (Node *)rm);
}

void nb_SendReadEventReq(struct NB_IOEventMsg * rm)
{
  nb_SendReadReq((struct NB_IOMsg *)rm);
}

/****/

void nb_SendWriteEventReq(struct NB_IOEventMsg * wm)
{
#ifdef DEBUG  
  if (!FD_ISSET(wm->fd, &NBCPriv.fds)) {
    TRACEF(1, ("Descriptor %d not set ready for nb_SendWriteReq\n", wm->fd));
    nb_SetFd(wm->fd);
  }
#endif
  addTail(&NBCore.addwritelist, (Node *)wm);
}

void nb_SendWriteReq(struct NB_IOMsg * wm)
{
  wm->res = 0;
  nb_SendWriteEventReq((struct NB_IOEventMsg *)wm);
}

/****/

void nb_SendReplyMsg(NB_Msg * m)
{
  assert(m->node.ln_Succ == 0);
  addTail(&NBCore.replylist, (Node *)m);
}

void nb_AbortTimeReq(struct NB_TimeMsg * msg)
{
  nb_AbortMsg((NB_Msg *)msg);
}

void nb_AbortReadEventReq(NB_IOEventMsg * msg)
{
  nb_AbortReadReq((NB_IOMsg *)msg);
}

void nb_AbortWriteEventReq(NB_IOEventMsg * msg)
{
  nb_AbortWriteReq((NB_IOMsg *)msg);
}

#endif /* __OPTIMIZE__ */


/***************************************************************************
 * Blocking write is useful in some situations
 */

#if NB_HAVE_BLOCKWRITE

#if 1
int nb_BlockWrite(int fd, const void * buf, int len)
{
  int rv;
  int l, n;

  rv = write(fd, buf, len);

  if (rv == len)
    return rv;

  if (rv < 0)
    if (errno == EAGAIN) rv = 0; /* EINTR impossible if in NONBLOCK mode */
    else return -1;
  
  n = fd + 1;
  buf += rv;
  l = len - rv;

  while (1) {
    fd_set wfds = { 0 };
  
    FD_SET(fd, &wfds);

    if (select(n, NULL, &wfds, NULL, NULL) < 0)
      if (errno == EINTR) continue;
      else return -1;
    
    /* write must be set if here */

    rv = write(fd, buf, l);

    if (rv == l)	return len;
    if (rv < 0)		return len - l;

    buf += rv;
    l -= rv;
  }
  /* NOT REACHED */
  return 0;
}

#else

int nb_BlockWrite(int fd, const void * buf, int len)
{
  int rv;
  ioctl(fd, FIONBIO, &ZERO); /* clear non-blocking IO */
  ioctl(fd, FIOASYNC, &ONE); /* set asynchronous IO */

  rv = write(fd, buf, len);
  
  ioctl(fd, FIONBIO, &ONE); /* set non-blocking IO */
  ioctl(fd, FIOASYNC, &ZERO); /* clear asynchronous IO */

  return rv;
}
#endif

#endif /* NB_HAVE_BLOCKWRITE */
