/*
 * $Id: nbcore.h,v 1.1 1996/06/29 17:10:55 too Exp too $
 *
 * Author: Tomi Ollila <too@nsdi.fi>
 *
 * 	Copyright (c) 1996 Network Solutions Development, Inc.
 * 	    All rights reserved
 *
 * Created: Wed Apr 24 20:16:36 1996 too
 * Last modified: Mon Sep  7 08:41:41 1998 too
 *
 * HISTORY 
 * $Log: nbcore.h,v $
 * Revision 1.1  1996/06/29 17:10:55  too
 * Initial revision
 *
 */

#ifndef NBCORE_H
#define NBCORE_H

/* Change default behaviour using compiler flags */

#ifndef NB_HIRES_TIMERS
#define NB_HIRES_TIMERS 0
#endif

/* if IOMSGBUFSIZ > 0, NB_IOMsg buffer includes the input buffer, but then
   data must always be read at the start of buffer... */
#ifndef NB_IOMSGBUFSIZ
#define NB_IOMSGBUFSIZ 0
#endif

#ifndef NB_HAVE_BLOCKWRITE
#define NB_HAVE_BLOCKWRITE 0
#endif


#ifndef LIST_H
#include "list.h"
#endif

#include <time.h>

#if NB_HIRES_TIMERS
extern struct timeval NB_TIMEVAL;
#define NB_TIME NB_TIMEVAL.tv_sec
#else
extern unsigned long  NB_TIME;
#endif  

#ifndef __sun__ /* problems w/ solaris. just now won't bother with it */
#ifdef DEBUG
#define close(x) nb_ChkClose(x)
#endif
#endif

/*
 * NBClists structure had to make public so that inline functions can access
 * the fields. The reason for private structure in `nbcore.c' is there so
 * that not all the modules including this file needs to include
 * <sys/types.h> (and perhaps some others too)
 */

extern struct NBCore {
  List	 addreadlist;
  List	 addwritelist;
  List	 timerlist;
  List	 replylist;
} NBCore;

/*
 * Head of message structure. Used additions come below.
 * i.e. all these structures are to be extended by user additions
 * w/ c++ this would work much better, but these works also w/ c.
 */

typedef struct NB_Msg
{
  Node		node;
  void	(*	func)(struct NB_Msg *);
} NB_Msg;

/* Spesific messages that have Msg structure in head (easier to use) */

typedef struct NB_InitMsg
{
  Node		node;
  int 	(*	func)(struct NB_Msg *);
} NB_InitMsg;


/*
 * NB_IOMsg is used for reading and writing to fd. buffer and len used.
 */
typedef struct NB_IOMsg
{
  Node		node;
  void	(*	func)(struct NB_IOMsg *);

  int		fd;
  int		len; /* length of buffer */
  int		res; /* data transferred from/to buffer */
#if NB_IOMSGBUFSIZ
  char		data[NB_IOMSGBUFSIZ];
#else  
  void *	buf;
#endif

} NB_IOMsg;

/*
 * when NB_IOEventMsg is used, only event indication is sent to the callee.
 */
typedef struct NB_IOEventMsg
{
  Node		node;
  void	(*	func)(struct NB_IOEventMsg *);

  int	fd;
  int	len;	/* len must be zero for event indications */
} NB_IOEventMsg;

/*
 * structure for timer message
 */
typedef struct NB_TimeMsg
{
  Node		node;
  void	(*	func)(struct NB_TimeMsg *);

  unsigned long	sec;
#if NB_HIRES_TIMERS
  long		usec;
#endif  
} NB_TimeMsg;


void nb_init(void);
int nb_start(void);
void nb_main(void);

void nb_SendTimeReq(NB_TimeMsg * tm);

void nb_AbortMsg(NB_Msg * msg);
void nb_AbortReadReq(NB_IOMsg * msg);
void nb_AbortWriteReq(NB_IOMsg * msg);

void nb_SetSignals(int sig1, ...);
void nb_SetSigHandler(void (*handler)(int, void *), void * data);

void nb_SetFd(int fd);
void nb_CloseFd(int fd);
void nb_ChkClose(int fd);

#if NB_HAVE_BLOCKWRITE
int nb_BlockWrite(int fd, const void * buf, int len);
#endif

#ifndef __OPTIMIZE__

void nb_SendReadReq(NB_IOMsg * rm);
void nb_SendReadEventReq(NB_IOEventMsg * rm);
void nb_SendWriteReq(NB_IOMsg * wm);
void nb_SendWriteEventReq(NB_IOEventMsg * wm);
void nb_SendReplyMsg(NB_Msg * m);

void nb_SendTimeReq1(NB_TimeMsg * tm, unsigned long sec);
void nb_SendTimeReq2(NB_TimeMsg * tm, unsigned long sec, long usec);

void nb_AbortTimeReq(struct NB_TimeMsg * msg);
void nb_AbortReadEventReq(NB_IOEventMsg * msg);
void nb_AbortWriteEventReq(NB_IOEventMsg * msg);

#else

#ifndef INLINE
#define INLINE extern __inline__
#endif

#if NB_HIRES_TIMERS

INLINE void nb_SendTimeReq1(struct NB_TimeMsg * tm, unsigned long sec)
{
  tm->sec = NB_TIMEVAL.tv_sec + sec;
  nb_SendTimeReq(tm);
}

void nb_SendTimeReq2(struct NB_TimeMsg * tm, unsigned long sec, long usec);

#else	/* not NB_HIRES_TIMERS */

INLINE void nb_SendTimeReq1(struct NB_TimeMsg * tm, unsigned long sec)
{
  tm->sec = NB_TIME + sec;
  nb_SendTimeReq(tm);
}

INLINE void nb_SendTimeReq2(struct NB_TimeMsg * tm,
			    unsigned long sec, long usec)
{
  tm->sec = NB_TIME + sec;
  nb_SendTimeReq(tm);
}

#endif

/****/

INLINE void nb_SendReadReq(struct NB_IOMsg * rm)
{
  addTail(&NBCore.addreadlist, (Node *)rm);
}

INLINE void nb_SendReadEventReq(struct NB_IOEventMsg * rm)
{
  addTail(&NBCore.addreadlist, (Node *)rm);
}

/****/

INLINE void nb_SendWriteReq(struct NB_IOMsg * wm)
{
  wm->res = 0;
  addTail(&NBCore.addwritelist, (Node *)wm);
}

INLINE void nb_SendWriteEventReq(struct NB_IOEventMsg * wm)
{
  addTail(&NBCore.addwritelist, (Node *)wm);
}

/****/

INLINE void nb_SendReplyMsg(NB_Msg * m)
{
  addTail(&NBCore.replylist, (Node *)m);
}

/****/

INLINE void nb_AbortTimeReq(struct NB_TimeMsg * msg)
{
  nb_AbortMsg((NB_Msg *)msg);
}

INLINE void nb_AbortReadEventReq(NB_IOEventMsg * msg)
{
  nb_AbortReadReq((NB_IOMsg *)msg);
}

INLINE void nb_AbortWriteEventReq(NB_IOEventMsg * msg)
{
  nb_AbortWriteReq((NB_IOMsg *)msg);
}

#endif /* NO_INLINE_FUNCTIONS */

#endif /* NBCORE_H */
