/*
 * $Id: prgutil.c $
 *
 * Author: Tomi Ollila <Tomi.Ollila@iki.fi>
 *
 * Created: Thu Mar 19 18:35:18 1998 too
 * Last modified: Wed Apr  1 16:25:19 1998 too
 *
 * HISTORY 
 * $Log: $
 */

/* stuff mainly ripped from nbcore -- since can be used separately */

#include "userlog.h"
#include "prgutil.h"

#define TRACEF(x,y);

#define NULL 0
#define TRUE 1
#define FALSE 0

static struct PRG
{
  PRG_Args * argl;	/* modules desire to have command line arguments */
  PRG_ExitMsg * eml;
} PRG;

void prg_AddArgList(struct PRG_Args * args)
{
  args->next = PRG.argl;
  PRG.argl = args;
}

void prg_AddExitMsg(struct PRG_ExitMsg * em)
{
  em->next = PRG.eml;
  PRG.eml = em;
}

int arrstrstr(const char * stringarray[], const char * string)
{
  int i;

  for (i = 0; stringarray[i]; i++)
    if (strcmp(stringarray[i], string) == 0)
      return i;
  return -1;
}

char * have_next_arg(int * argcp, char *** argvp)
{
  if (*argcp == 0)
    return NULL;
  
  (*argcp)--;
  return *((*argvp)++);
}

int prg_ArgParse(int * argcp, char *** argvp)
{
  int i = 0; /* would be uninitialized if PRG.argl is empty */
  char * arg;
  struct PRG_Args * argl;
  struct PRG_Args * narg;
    
  /* first, tell program name to everyone */
  for (argl = PRG.argl; argl; argl = argl->next)
    if ((i = argl->argfunc(argl, -2, argcp, argvp)))
      return i;

  (*argcp)--;
  (*argvp)++;

  while ((arg = have_next_arg(argcp, argvp))) {
    /*
     * check switch condition
     */
    TRACEF(10, ("Arg: %s\n", arg));
    if (*arg == '-') {

      if (strcmp(arg, "--") == 0) /* `--' -> no more switches */
	break;
      
      for (argl = PRG.argl; argl; argl = argl->next)
	if ((i = arrstrstr(argl->argtable, arg)) >= 0)
	  if ((i = argl->argfunc(argl, i, argcp, argvp)))
	    return i;
	  else
	    break;

      if (i < 0) {
	userlog(-1, "`%s' unknown switch.\n", arg);
	return 1;
      }
    }
    else {
      (*argcp)++;
      (*argvp)--;
      break;
    }
  }
  /* no more switches, tell rest argument array to everyone */
  /* argl structure may be deleted when here called */
  for (argl = PRG.argl; argl; argl = narg)
  {
    narg = argl->next;
    if ((i = argl->argfunc(argl, -1, argcp, argvp)))
      return i;
  }

  return 0;
}

/*
 * TODO: do a single item finalization routine.
 */

void prg_Exit(int rv)
{
  struct PRG_ExitMsg * em;

  for (em = PRG.eml; em; em = em->next)
    em->exitfunc(em);
  exit(rv);
}

char * need_next_arg(int * argcp, char *** argvp)
{
  char * p = have_next_arg(argcp, argvp);

  if (p == NULL)
  {
    userlog(-1, "Argument(s) required after `%s'", (*argvp)[-1]);
    prg_Exit(-1);
  }
  return p;
}
