/*
 * $Id: userlog.c,v 1.9 1996/10/24 13:31:41 too Exp $
 *
 * Author: Tomi Ollila <too@iki.fi>
 *
 * Last modified: Sat Mar 21 17:14:09 1998 too
 *
 * HISTORY 
 * $Log: userlog.c,v $
 */

/*
 * Lots of stuff is outcommented for this example, since all necessary
 * stuff for them to work is not available.
 */

/* edit this to your personal needs (for example a daemon may need to write
 * to a log, a text based client to screen, a curses based client to a window
 * and a GUI client to GUI window
 */

#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#include "userlog.h"

#include <sys/types.h>
#include <sys/stat.h>

extern int errno;
extern const char * const sys_errlist[];

static char NULSTR[] = "";

static int idummy() { return 0; }

static struct USERLOG {
  char *	filename;
  int		(*datefunc)(char *);
  char *	ident;
  int		priority;
 int		nlmode;
} USERLOG = { NULL, idummy, NULSTR, 7, 1 };

#if 0 /* outcommented since not tested */
static void freeRes(void)
{
  if (USERLOG.filename) {
    free(USERLOG.filename);
    USERLOG.filename = NULL;
  }
  if (USERLOG.ident != NULSTR) {
#ifdef COPY_UL_IDENT    
    free(USERLOG.ident);
#endif    
    USERLOG.ident = NULSTR;
  }
}
#endif

const char * setIdent(char * ident)
{
  if (ident != NULL) {
#ifdef COPY_UL_IDENT
    if (USERLOG.ident != NULSTR)
      free(USERLOG.ident);
    USERLOG.ident = strdup(ident);
#else
    USERLOG.ident = ident;
#endif
  }
  return USERLOG.ident;
}

#if 0  /* outcommented since not tested */
int openuserlog(char * filename, char * ident, unsigned int priority,
		int (*datefunc)(char *))
{
  int (*odf)(char *) = USERLOG.datefunc;

  USERLOG.datefunc = datefunc;
  USERLOG.filename = (filename)? strdup(filename): NULL;
  setIdent(ident);
  USERLOG.nlmode = 1;

  USERLOG.priority = 7;
  if (userlogf(6, "*** Log started.\n") == 0)
    USERLOG.priority = (priority > 7)? 7: priority;
  else {
    freeRes();
    USERLOG.datefunc = odf;
    return -1;
  }
  return 0;
}
#endif

#if 0  /* outcommented since not tested */
void closeuserlog()
{
  USERLOG.priority = 7;
  userlogf(6, "*** Log stopped.\n");

  freeRes();
}
#endif

#if 0   /* outcommented since not tested */  

#ifndef LOGFILESIZE
#define LOGFILESIZE 100000
#endif

static void cycle_userlog(char * buf)
{
  struct stat st;

  if (stat(USERLOG.filename, &st) >= 0)
  {
    if (st.st_size > LOGFILESIZE) {
      int fd = do_cycling(buf, USERLOG.filename, -1, 4);
      close(fd);
    }
  }
}
#endif

static char * pri[] =
{
  "fatal", "emerg", "alert", "crit", "err",
  "warning", "notice", "info", "debug"
};


int userlog(int priority, char * format, ...)
{
  va_list ap;
  int rv;
  
  va_start(ap, format);
  rv = vuserlog(priority, format, (char *)ap);
  va_end(ap);
  return rv;
}

/* ok, tm userlog funktio toimii vain, jos %m on viimeisena (jos ollenkaan) */
int vuserlog(int priority, char * format, char * ap)
{
  char buf[16384];
  int i, fd;
  char * ptr;
  char * start;

  if (USERLOG.priority < priority)
    return 0;

#if 0  /* outcommented since not tested */  
  if (USERLOG.filename == NULL)
    fd = 2;	/* output to console */
  else {
    cycle_userlog(buf);

    if ((fd = open(USERLOG.filename, O_WRONLY|O_CREAT|O_APPEND, 0664)) < 0)
      return -1;
  }
#else
  fd = 2;
#endif

  if ((ptr = strstr(format, "%m")) && (ptr == format || *(ptr - 1) != '%')) {
    start = buf + strlen(format) + 1;
    strcpy(buf, format);
    ptr+= (buf- format);
  }
  else
    start = buf;

  if (USERLOG.nlmode) {
    i = USERLOG.datefunc(start);
    if (i)
      start[i++] = ' ';
    sprintf(start + i, "%s: %s: ", USERLOG.ident, pri[priority + 1]);
    i+= strlen(start + i);
  }
  else
    i = 0;

  if (ptr) {
    if (ptr > buf) {
      *ptr = '\0';
      vsprintf(start + i, buf, (va_list)ap);
      i += strlen(start + i);
      ptr += 2;
    }
    strcpy(start + i, sys_errlist[errno]);
    i += strlen(start + i);
  }
  else
    ptr = format;

  vsprintf(start + i, ptr, (va_list)ap);

  i = strlen(start + i) + i;
  write(fd, start, i);

  USERLOG.nlmode = (start[i-1] == '\n')? 1: 0;

  if (USERLOG.filename)
    close(fd);
  return 0;
}
